package GO;
use strict;
use Symaxx2::MySubs;
use Text::Wrap qw(wrap $columns $huge);;
use Symaxx2::MaximaSymbols;
# TYPE: What is this
# - LIST : DATA contains list of arguments
# - ATOM : DATA1 contains argument as string

# todo:NewGoList - remove command?


#Lisplink is the pointer to the Lisp data structure, to which this graphic object corresponds to.
#It is not always identical with Lispobject: For example in fractions, the numerator and denominator
#are artificially created - they are UNREAL. Changing those objects would not change the real 
#data structure itself.
#But since the artificial objects use pointers to the real contents in most cases, the objects on lower
#levels can be modified.
#If an artificial object is selected, it will give a pointer to the next higher real object.
#That's what LISPLINK is good for.

# A word on style (STYLE parameter for new): 
# If a new graphobject has a STYLE parameter, then it will create a coloured background, before any child objects are drawn.
# For the child objects, it is completely unnecessary to repeat the Style (since the background is coloured anyway). This would be a waste of computing power, avoid it.
# Of course, if we return another object in this place, then STYLE has to be handed on.
# Summary: return new...(STYLE => $Style): Correct
# but push @List,new(STYLE => $Style): Wrong, avoid this
sub NewGOList{
    # create an arbitrary list of objects, which are to be lined up in X-or Y direction
    my $self={};
    bless $self;	
    $::GOCount++; # Memory debug code

    my %P=@_;
    my $GOList=$P{'GOLIST'} or die;
    my $LispLink=$P{'LISPLINK'} or die;
    my $Style=$P{'STYLE'};
    my $Type=$P{'TYPE'} or die;
    my $Command=$P{'COMMAND'}; # optional
    $self->{'ARGS'}=$GOList;
    $self->{'TYPE'}=$Type;
    if ($Command){
	$self->{'COMMAND'}=$Command;
    };
    $self->{'LISPLINK'}=$LispLink; 
    if ($Style){$self->{'STYLE'}=$Style;};
    return $self;
};
sub NewPicture{
    my $self={};
    bless $self;  
    $::GOCount++; # Memory debug code
    my %P=@_;
    my $LispLink=$P{LISPLINK};
    if ($LispLink){$self->{LISPLINK}=$LispLink};
    # The UI-manager maintains a list of pictures. It uses the ID and
    # the identifier as keys. The GO must know them. Epsfile is optional.

    # Handle is created here.
    my $EPSPathname=$P{EPSFILE} or die;  # ??????;
    my $Scale=$P{SCALE};
    my $Handle=$P{HANDLE}; die if defined $Handle;
    if ($EPSPathname){
	$self->{EPSFILE}=$EPSPathname;
    };
    $self->{TYPE}='PICTURE';

    # Handle creation
    $Handle=$EPSPathname.$Scale;

    $self->{HANDLE}=$Handle;

    # load the photo to memory here. 
    my $Photo=$::UIManager->GetPicture(HANDLE => $Handle,EPSFILE => $EPSPathname,SCALE =>$Scale);
    if ($Photo){
	return $self;
    } else {return undef;}
};
sub NewInBrackets{
    my $self={};
    bless $self;    $::GOCount++; # Memory debug code
    my %P=@_;
    my $LispLink=$P{'LISPLINK'} or die;
    my $Lispobject=$P{'LISPOBJECT'};
    my $Graphobject=$P{'GRAPHOBJECT'};
    my $Bracket=$P{'BRACKETS'} or die;
    my $Style=$P{'STYLE'};

    die unless (($Bracket eq 'ROUNDBRACKETS')||($Bracket eq 'SQUAREBRACKETS')||($Bracket eq 'VERTBARS')||($Bracket eq 'INTEGRATE')||($Bracket eq 'SUM'));
    if ($Lispobject){
	if (!($Lispobject->QueryUnreal)){$LispLink=$Lispobject;};
	die "$Lispobject is no lisp object" unless $Lispobject =~/LispList/;
	$self->{'ARGS'}=[GO::New('LISPOBJECT'=>$Lispobject,
				 'PARENTOPERATOR'=>$Bracket,
				 'LISPLINK'=>$LispLink)];
    } elsif ($Graphobject){
	$self->{'ARGS'}=[$Graphobject];
    } else {die;}
    $self->{'TYPE'}=$Bracket;
    $self->{'LISPLINK'}=$LispLink;   
    if ($Style){$self->{'STYLE'}=$Style;};
    return $self;
};	
sub NewFromString{
    my $self={};
    bless $self;    $::GOCount++; # Memory debug code
    my %P=@_;
    my $NewString=$P{'STRING'}; die unless defined $NewString;
    my $LispLink=$P{'LISPLINK'}; die unless defined $LispLink;
    my $ParOp=$P{'PARENTOPERATOR'}; # not necessarily defined
    my $PosInParent=$P{'POSINPARENT'}; # not necessarily defined
    my $Background=$P{'BACKGROUND'}; # not necessarily defined
    my $Style=$P{'STYLE'};
    my $Wrap=$P{'WRAP'}; 
    my $Font=$P{FONT}; 

    my $Brackets=0;
    # Conditions, when this has to be put into brackets:
    # negative? 
    if ($ParOp && $NewString=~/^\-(\d+)/){
	# a*-3
	if (($ParOp eq 'MTIMES')&&($PosInParent>0)){$Brackets=1;}
	# -3^a
	if (($ParOp eq 'MEXPT') && ($PosInParent==0)){$Brackets=1;}
	if ($Brackets){
	    return GO::NewInBrackets('LISPOBJECT'=>
				   LispList::NewUnreal('STRING' => $NewString),
				     'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS');
	}; # if brackets needed
    }; # if negative number
    $self->{'TYPE'} = 'ATOM';
    if ($Wrap){$self->{'WRAP'} = $Wrap;}
    $self->{'LISPLINK'}=$LispLink;
    
    # check for units
    if ($::MaximaUseUnits && defined $::Units{$NewString}){
	$Style='UNIT';
	if ($::Units{$NewString}){
	    $NewString=$::Units{$NewString}; # translate NEWTON => N etc
	};
    };
    $self->{'DATA1'}=$NewString;
    if ($Style){$self->{'STYLE'}=$Style;};
    if ($Background){$self->{'BACKGROUND'}=$Background;};
    if ($Font){$self->{FONT}=$Font;};
    return $self;
};
sub NewSeparator{
    my $self={};
    bless $self;    $::GOCount++; # Memory debug code
    my $Sep=shift or die;
    my $LispLink=shift or die;
    $self->{'ARG1'}=$Sep;
    $self->{'TYPE'}='SEPARATOR'; 
    $self->{'LISPLINK'}=$LispLink;
    # no style. What would we need that for?
   return $self;
};
sub INeedBrackets{
    my $self=shift or die;
    my %P=@_;
    my $Command=$self->{'COMMAND'}; return unless $Command;
    my $ParOp=$P{'PARENTOPERATOR'}; return unless $ParOp;
    my $Lispobject=$P{'LISPOBJECT'} or die;
    my $LOInfo=$Lispobject->QueryGeneralInfo;
    my $PosInParent=$P{'POSINPARENT'}; # not necessarily defined
    my $Brackets=0;
    my $CommandCopy=$Command;
    
    if ($LOInfo->{'COMMANDUSEREXP'}||$LOInfo->{'COMMANDSYSTEMEXP'}){
	$CommandCopy='FUNCTION';
    };
    
    my $ParOpLookupCommand=$ParOp;
    if ($ParOp eq 'MEXPT' && defined $PosInParent){$ParOpLookupCommand.=$PosInParent;};
    
    my $ParOpBP=0; # 'BindPower'
    if ($::GraphBindPowers{$ParOpLookupCommand}){$ParOpBP=$::GraphBindPowers{$ParOpLookupCommand};};
    my $ThisBP=0; # 'BindPower'
    if ($::GraphBindPowers{$CommandCopy}){$ThisBP=$::GraphBindPowers{$CommandCopy};};
#	print "$ParOpBP $ThisBP $ParOpLookupCommand $CommandCopy\n";
    if ($ParOpBP<9999){
	if ($ThisBP<=$ParOpBP){$Brackets=1;};
    };
    if ($ParOp eq 'MEXPT' && ($Command eq 'MEXPT')){$Brackets=1;}
    if ($ParOpLookupCommand eq 'MEXPT0' && ($Command eq 'FRACTION'||$Command eq 'QUOTIENT'||$Command eq 'RAT')){$Brackets=1;}
    # this is a special rule for (a^b)^c and a^(b^c): use always brackets, since the graphical 
    # representation is ambiguous (at least for me...)
    # I'm not sure if the next one makes sense. a^(b/c) produces brackets, but a^((b/c)+d) not.
#	if ($ParOp eq 'MEXPT' && ($Command eq 'FRACTION')){$Brackets=1;}
    return $Brackets;
};
sub New{
    my %P=@_;
#    print "New GO: $self\n";
    # Here we're getting a bunch of arguments. We got to decide what to make out of them.
    # If no rule applies, we make a lisp list: (a,b,c,...).
    # If there is only a single argument, and it's a string, then we make an atom instead.
    my $ParOp=$P{'PARENTOPERATOR'};
    my $PosInParent=$P{'POSINPARENT'}; # not necessarily defined
    my $LispLink=$P{'LISPLINK'} or die;
    my $Lispobject=$P{'LISPOBJECT'} or die;
    my $LOInfo=$Lispobject->QueryGeneralInfo or die;
    my $Style=$P{'STYLE'};
    my $Wrap=$P{'WRAP'};
    # When you click on this object, should you be able to edit the LispObject, or should some higher lispobject be 
    # selected (as given in LispLink)?
    if (!$LOInfo->{'UNREAL'}){
	$LispLink=$Lispobject;
    };

    #####################################################################
    # 1) Test, if we turn this to atom    
    #####################################################################
    
    # this is returned instead of $self.
    my $AtomData=$LOInfo->{'ATOMDATA'};
    if (defined $AtomData){
	return GO::NewFromString('STRING' => $AtomData,
				 'PARENTOPERATOR'=>$ParOp,
				 'POSINPARENT'=>$PosInParent,
				 'LISPLINK'=>$LispLink,
				 'STYLE' => $Style,
				 'WRAP' => $Wrap);
    }
    
    my $Command=$LOInfo->{'COMMAND'};   
    
    # we create the object body only here to make it faster (Atom needs none, returns another object).
    my $self={}; 
    
    # We can prevent, that an object is selected directly.
    # This is used for the SEL command and RED, GREEN,...: The frame is as big as
    # the contents, so there is no way to select the argument.

    if ($Command){
	my @ArgList=@{$Lispobject->QueryArgs} or die;
	
	# debug code
	my $Elem;
	foreach $Elem(@ArgList){
	    die "-$Elem-" unless ref($Elem) eq 'LispList';
	};
	#end debug code
	die "$Lispobject is no lisp object" unless ref($Lispobject) eq 'LispList';
	
	#    print "New GO: ".$Lispobject->QueryDebugString."\n";
	
	# UE means User Expression (starting with $)
	my $UE=$LOInfo->{'COMMANDUSEREXP'}; 
	# SE means System Expression (starting with %)
	my $SE=$LOInfo->{'COMMANDSYSTEMEXP'};
	# BE means Blank expression (no character in front)
	my $BE=!($UE || $SE); 
	my $Done=0;
	
	if (0){ # disable smart drawing due to the use of DISPFORM
	    # Well, maybe we'll use it anyway...
	    # OK, we'll have to 
	if ($BE){
	    #####################################################################
	    # Test, if we can transform this into a simpler expression.
	    # This has to be done before brackets handling.	
	    # The following rules do not process all 'MTIMES' or 'MEXPT'-commands!
	    #####################################################################
	    if ($Command eq 'MTIMES'){
		my @Numerator;my @Denominator;
		# Go through all the arguments of this MTIMES and put them either into the 
		# numerator or the denominator.
		my $Count=0;for ($Count=1;$Count<=$#ArgList;$Count++){
		    my $InvertedArg=$ArgList[$Count]->QueryInverted;
		    if ($InvertedArg){
			push @Denominator,$InvertedArg; die unless $InvertedArg=~/LispList/;
		    } else {
			push @Numerator,$ArgList[$Count];
		    };		
		}; # for $Count
		# if a denominator exists, then the type changes to fraction
		if ($#Denominator>=0){
		    if ($#Numerator==-1){push @Numerator,LispList::NewFromString(STRING => '1');};
		    $Command='FRACTION';
		    my $DenomObj=LispList::NewUnreal('OPERATOR' => 'MTIMES','DATA' => \@Denominator); 
		    my $NumObj=LispList::NewUnreal('OPERATOR' => 'MTIMES','DATA' => \@Numerator);
		    $self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>$NumObj,
						 'PARENTOPERATOR' => 'FRACTION',
						 'LISPLINK'=>$LispLink) or die;
		    $self->{'ARGS'}->[1]=GO::New('LISPOBJECT'=>$DenomObj,
						 'PARENTOPERATOR' => 'FRACTION',
						 'LISPLINK'=>$LispLink) or die;
		    $Done=1;
		}; # if turned to fraction
	    } # if MTIMES
	    elsif ($Command eq 'MEXPT'){		
		# look out for x^-a, x^-2 etc
		my $InvertedArg=$Lispobject->QueryInverted;	
		if ($InvertedArg){
		    $Command='FRACTION';
		    $self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>LispList::NewUnreal('STRING' => '1'),
						 'PARENTOPERATOR' => 'FRACTION','LISPLINK'=>$LispLink) or die;
		    $self->{'ARGS'}->[1]=GO::New('LISPOBJECT'=>$InvertedArg,
						 'PARENTOPERATOR' => 'FRACTION','LISPLINK'=>$LispLink) or die;
		    $Done=1;
		}; # if inversion works
		# check for square root
		my $Arg2Inverted=$ArgList[2]->QueryInverted;
		if ($Arg2Inverted){
		    my $String=$Arg2Inverted->QueryAtomData;
		    if ($String && $String eq '2'){
			$Command='SQRT';
			$self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>$ArgList[1],
						     'PARENTOPERATOR' => 'SQRT','LISPLINK'=>$LispLink) 
			    or die;		
			$Done=1;
		    }; # if to the power of 1/2
		}; # if to the power of 1/x
	    }; # if MEXPT	
	};
    }; # disable smart drawing
	# if Non-Userexpressions (no $ as first character)
	if (!$Done){
	    #####################################################################
	    # 3) Now come straightforward rules for each function. Those rules
	    # must process their respective command, because of the 
	    # if ... elsif ... structure.
	    # For better runtime performance, often used expressions should
	    # be higher in the list.
	    #####################################################################
	    if (($Command eq 'MTIMES' || $Command eq 'MPLUS' || $Command eq 'MMINUS' 
		 || $Command eq 'MEQUAL' || $Command eq 'MNOTEQUAL'|| $Command eq 'MLESSP'|| $Command eq 'MGREATERP'
		 || $Command eq 'MLEQP'|| $Command eq 'MGEQP'|| $Command eq 'MLIST' 
		 || $Command eq 'MPROG' || $Command eq 'MPROGN' || $Command eq 'MSETQ' || $Command eq 'MDEFINE' || $Command eq 'MDEFMACRO' || $Command eq 'MAND' || $Command eq 'MOR') && $BE
		|| ($Command eq 'MATRIX' || $Command eq 'YLIST') && $UE){
		# Generate argument GOs	
		if ($Command eq 'MMINUS' && $#ArgList !=1){die 'MMINUS should have only one arg?! Back to basics!';};
		my @GOList=();
		my $Count;for ($Count=1;$Count<=$#ArgList;$Count++){
		    my $OrigElem=$ArgList[$Count] or die;
		    die "-$Command-$OrigElem-".join(" ",@ArgList) unless $OrigElem=~/LispList/;
		    my $NewElem;my $Separator='';
		    # a+-b => a - b
		    if ($Command eq 'MPLUS' && $OrigElem->QueryNegated){
			$NewElem=$OrigElem->QueryNegated;
			$Separator='-';
		    } elsif ($Command eq 'MTIMES'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='*';}
		    } elsif ($Command eq 'MPLUS'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='+';}
		    } elsif ($Command eq 'MMINUS'){
			$NewElem=$OrigElem;$Separator='-';
		    } elsif ($Command eq 'MEQUAL'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='=';}
		    } elsif ($Command eq 'MNOTEQUAL'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='#';}
		    } elsif ($Command eq 'MLESSP'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='<';}
		    } elsif ($Command eq 'MGREATERP'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='>';}
		    } elsif ($Command eq 'MLEQP'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='<=';}
		    } elsif ($Command eq 'MGEQP'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='>=';}
		    } elsif ($Command eq 'MLIST'){
			$NewElem=$OrigElem;if ($Count>1){$Separator=',';}
		    } elsif ($Command eq 'MSETQ'){
			$NewElem=$OrigElem;if ($Count>1){$Separator=':';}
		    } elsif ($Command eq 'MDEFINE'){
			$NewElem=$OrigElem;if ($Count>1){$Separator=':=';}
		    } elsif ($Command eq 'MDEFMACRO'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='::=';}
		    } elsif ($Command eq 'MOR'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='|';}
		    } elsif ($Command eq 'MAND'){
			$NewElem=$OrigElem;if ($Count>1){$Separator='&';}
		    } elsif ($Command eq 'MPROGN'){
			$NewElem=$OrigElem;if ($Count>1){$Separator=',';}
		    } else {
			$NewElem=$OrigElem;
		    };
		    if ($Separator){
			push @GOList,&NewSeparator($Separator,$LispLink);
		    }; # if separator
		    die unless $NewElem=~/LispList/;
		    my $NewObj=GO::New('LISPOBJECT' => $NewElem,
				       'PARENTOPERATOR' => $Command,
				       'POSINPARENT' => $Count,
				       'LISPLINK'=>$LispLink);
		    push @GOList,$NewObj;
		}; # foreach $Elem
		my $Listtype='XLIST'; if ($Command eq 'MATRIX'||$Command eq 'YLIST'|| $Command eq 'MPROG'){$Listtype='YLIST';}
		my $Object;
		if ($Command eq 'MLIST'|| $Command eq 'MATRIX'|| $Command eq 'MPROGN'|| $Command eq 'MPROG'){
		    my $List=&NewGOList('TYPE' => $Listtype,'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList);
		    my $BracketsType='SQUAREBRACKETS';
		    if($Command eq 'MPROGN'){$BracketsType='ROUNDBRACKETS';};
		    $Object=GO::NewInBrackets('GRAPHOBJECT'=>$List,'LISPLINK'=>$LispLink,
					      'BRACKETS'=> $BracketsType,'STYLE' => $Style);		
		}elsif ($Command eq 'YLIST'){
		    my @GOList2=&NewGOList('TYPE' => $Listtype,'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList,'STYLE' => $Style);		    
		    
		    push @GOList2,&NewSeparator(' ',$LispLink);
		    unshift @GOList2,&NewSeparator(' ',$LispLink);
		    return &NewGOList('TYPE' => 'XLIST','COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList2,'STYLE' => $Style);
		} else {	
		    $Object=&NewGOList('TYPE' => $Listtype,'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList,'STYLE' => $Style);
		};
		if ($Object->INeedBrackets('LISPOBJECT' => $Lispobject,'PARENTOPERATOR' => $ParOp,
					   'POSINPARENT' => $PosInParent)){
		    # xthe style is managed by the bracket object instead.
		    delete $Object->{'STYLE'};
		    return GO::NewInBrackets('GRAPHOBJECT'=>$Object,'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS','STYLE' => $Style);
		} else {return $Object;}
	    } elsif (($Command eq 'RAT'|| $Command eq 'MQUOTIENT') && $BE && $#ArgList == 2){
		# (rational) fraction
		$self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>$ArgList[1],'PARENTOPERATOR' => 'FRACTION',
					     'LISPLINK'=>$LispLink) or die; # No artificial object -> Lispobject
		$self->{'ARGS'}->[1]=GO::New('LISPOBJECT'=>$ArgList[2],'PARENTOPERATOR' => 'FRACTION',
					     'LISPLINK'=>$LispLink) or die;
		$Command='FRACTION';
	    } elsif ($Command eq 'MABS' && $BE && $#ArgList==1){
		my $NewObj=GO::New('LISPOBJECT' => $ArgList[1],
				   'PARENTOPERATOR' => 'FUNCTION',
				   'POSINPARENT' => 1,
				   'LISPLINK'=>$LispLink);
		# This will never need brackets.
		return GO::NewInBrackets('GRAPHOBJECT'=>$NewObj,'LISPLINK'=>$LispLink,
					 'BRACKETS'=> 'VERTBARS','STYLE' => $Style);
	    } elsif ($Command eq 'MQUOTE' && $#ArgList == 1 && $BE){
		my @GOList=(
			    &NewSeparator("'",$LispLink),
			  GO::New('LISPOBJECT' => $ArgList[1],
				  'PARENTOPERATOR' => $Command,
				  'POSINPARENT' => 1,
				  'LISPLINK'=>$LispLink));
		return &NewGOList('TYPE' => 'XLIST','COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList);
	    } elsif ($Command eq 'SQRT' && $#ArgList == 1 && $SE){
		$self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>$ArgList[1],'PARENTOPERATOR' => 'SQRT','LISPLINK'=>$LispLink) or die; 
	    } elsif ($Command eq 'AT' && $#ArgList == 2 && $UE){
		$self->{'ARGS'}->[0]=GO::New('LISPOBJECT'=>$ArgList[1],'PARENTOPERATOR' => 'AT0','LISPLINK'=>$LispLink) or die; 
		$self->{'ARGS'}->[1]=GO::New('LISPOBJECT'=>$ArgList[2],'PARENTOPERATOR' => 'AT1','LISPLINK'=>$LispLink) or die; 
	    } elsif ($Command eq 'MQAPPLY' && $BE){ 
		my @OuterGOList=(); # Arg1: Expression Arg2:[a,b,c,...] for example sin(x)[3], (a+b)[3,4]
		my @InnerGOList=(); # a,b,c or 3 or 3,4
		my $NewObj1=GO::New('LISPOBJECT' => $ArgList[1],
				    'PARENTOPERATOR' => 'FUNCTION',
				    'POSINPARENT' => 1,
				    'LISPLINK'=>$LispLink);
		if ($NewObj1->INeedBrackets('LISPOBJECT' => $ArgList[1],'PARENTOPERATOR' => 'MQAPPLY',
					    'POSINPARENT' => 1)){
		    $NewObj1=GO::NewInBrackets('GRAPHOBJECT'=>$NewObj1,'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS');
		}; # if first arg needs brackets (for example (REF[3]+REF[4])[1]
		push @OuterGOList,$NewObj1; #function name
		my $Count;for ($Count=2;$Count<=$#ArgList;$Count++){
		    my $Elem=$ArgList[$Count] or die;
		    my $Separator='';
		    die "-$Command-$Elem-".join(" ",@ArgList) unless $Elem=~/LispList/;
		    if ($Count>2){$Separator=',';}
		    if ($Separator){push @InnerGOList,&NewSeparator($Separator,$LispLink);};
		    my $NewObj=GO::New('LISPOBJECT' => $Elem,
				       'PARENTOPERATOR' => 'FUNCTION', # this behaves as a function: [a,b,c] is 'almost' fun(a,b,c)
				       'POSINPARENT' => $Count,
				       'LISPLINK'=>$LispLink);
		    push @InnerGOList,$NewObj;
		}; # foreach $Elem
		my $InnerXList=&NewGOList('TYPE' => 'XLIST','COMMAND' => 'PARTOF','LISPLINK'=>$LispLink,'GOLIST'=>\@InnerGOList);
		
		my $Obj2InBrackets=GO::NewInBrackets('GRAPHOBJECT'=>$InnerXList,'LISPLINK'=>$LispLink,
						     'BRACKETS'=> 'SQUAREBRACKETS');
		
		return &NewGOList('TYPE' => 'XLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>[$NewObj1,$Obj2InBrackets],'STYLE' => $Style);
	    } elsif ($Command eq 'MEXPT' && $BE && $#ArgList==2){  # a ^ b
		my $Lower=GO::New('LISPOBJECT' => $ArgList[1],
				  'PARENTOPERATOR' => $Command,
				  'POSINPARENT' => 0,
				  'LISPLINK'=>$LispLink);  
		my $Upper=GO::New('LISPOBJECT' => $ArgList[2],
				  'PARENTOPERATOR' => $Command,
				  'POSINPARENT' => 1,
				  'LISPLINK'=>$LispLink); 
		$self->{'ARGS'}=[$Lower,$Upper];
	    } elsif ((($Command eq 'INTEGRATE'  && ($#ArgList==2 || $#ArgList==4))|| $Command eq 'SUM' && $#ArgList==4) && ($UE || $SE)){
		my @YList;
		if ($#ArgList==4){
		    my $Upper=GO::New('LISPOBJECT' => $ArgList[4],
				      'LISPLINK'=>$LispLink); 
		    push @YList,$Upper;
		};
		my @MainArgList=();
		push @MainArgList,GO::New('LISPOBJECT' => $ArgList[1],
					  'LISPLINK'=>$LispLink);
		if ($Command eq 'INTEGRATE'){
		    push @MainArgList,GO::NewFromString('STRING'=>' d ',
							'LISPLINK'=>$LispLink); 
		    push @MainArgList,GO::New('LISPOBJECT' => $ArgList[2],
					      'LISPLINK'=>$LispLink);
		} 
		my $MainArg=&NewGOList('TYPE' => 'XLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@MainArgList);
		push @YList,GO::NewInBrackets('GRAPHOBJECT'=>$MainArg,'LISPLINK'=>$LispLink,'BRACKETS'=> $Command);
		if ($Command eq 'INTEGRATE'){
		    if ($#ArgList ==4){
			my $Lower=GO::New('LISPOBJECT' => $ArgList[3],
					  'LISPLINK'=>$LispLink); 
			push @YList,$Lower;
		    };
		} elsif ($Command eq 'SUM'){
		    my @LowerXList=();
		    push @LowerXList,GO::New('LISPOBJECT' => $ArgList[2],
					     'LISPLINK'=>$LispLink);
		    push @LowerXList,GO::NewFromString('STRING'=>'=',
						       'LISPLINK'=>$LispLink); 		
		    push @LowerXList,GO::New('LISPOBJECT' => $ArgList[3],
					     'LISPLINK'=>$LispLink); 
		    my $Lower=&NewGOList('TYPE' => 'XLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@LowerXList);
		    push @YList,$Lower;
		} else {die;}; # if sum
		return &NewGOList('TYPE' => 'YLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@YList,'STYLE' => $Style);
	    } elsif ((($Command eq 'DIFF' && $UE)|| ($Command eq 'DERIVATIVE' && $SE)) && ($#ArgList==2 || $#ArgList==3)){
		my @UpperXList;
		if ($#ArgList==2 ||$ArgList[3]->QueryGeneralInfo->{'ATOMDATA'} eq '1'){
		    push @UpperXList,GO::New('LISPOBJECT' => LispList::NewUnreal('STRING' => 'd '),
					     'LISPLINK'=>$LispLink);
		} else {
		    push @UpperXList,GO::New('LISPOBJECT' => LispList::NewUnreal
					     ('OPERATOR'=>'MEXPT',
					      'DATA'=>[LispList::NewUnreal('STRING'=>'d'),$ArgList[3]]),
					     'LISPLINK'=>$LispLink);
		    push @UpperXList,&NewSeparator(' ',$LispLink);
		    
		}; # d or d^n in numerator	
		push @UpperXList,GO::New('LISPOBJECT' =>$ArgList[1],'LISPLINK'=>$LispLink);
		
		$self->{'ARGS'}->[0]=GO::NewGOList('TYPE' => 'XLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@UpperXList);
		# Denominator
		my @LowerXList=(GO::New('LISPOBJECT' => LispList::NewUnreal('STRING' => 'd '),
					'LISPLINK'=>$LispLink));
		if ($#ArgList==2 ||$ArgList[3]->QueryGeneralInfo->{'ATOMDATA'} eq '1'){
		    push @LowerXList,GO::New('LISPOBJECT' =>$ArgList[2],'LISPLINK'=>$LispLink);
		} else {
		    # d x^n
		    push @LowerXList,GO::New('LISPOBJECT' => LispList::NewUnreal
					     ('OPERATOR'=>'MEXPT','DATA'=>[$ArgList[2],$ArgList[3]]),
					     'LISPLINK'=>$LispLink);
		}; # d or d^n in numerator	
		$self->{'ARGS'}->[1]=GO::NewGOList('TYPE' => 'XLIST', 'COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@LowerXList);
		$Command='FRACTION';
	    } elsif ($Command eq 'RED'||$Command eq 'GREEN'||$Command eq 'BLUE' ||$Command eq 'GRAY'){
		# Fun functions for color display :-)
		my @GOList=();
		push @GOList,&NewSeparator(' ',$LispLink);
		my $Count;for ($Count=1;$Count<=$#ArgList;$Count++){
		    my $Separator;
		    if ($Count>1){$Separator=' ';}
		    if ($Separator){
			push @GOList,&NewSeparator($Separator,$LispLink);
		    }; # if separator
		    my $NewObj=GO::New('LISPOBJECT' => $ArgList[$Count],
				       'PARENTOPERATOR' => 'DISPLAY',
				       'POSINPARENT' => $Count,
				       'LISPLINK'=>$LispLink);
		    push @GOList,$NewObj;	
		}; # for count
		push @GOList,&NewSeparator(' ',$LispLink);
		# this will never need brackets.
		my $GO=&NewGOList('TYPE' => 'XLIST','COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList,'STYLE'=>$Command);	
		return $GO;
	    } elsif (
		     ($Command eq 'TEXT'||
		      $Command eq 'HEADING1'||
		      $Command eq 'HEADING2' ||
		      $Command eq 'HEADING3'
		      ) && $#ArgList==1 
		     && defined($ArgList[1]->QueryGeneralInfo->{ATOMDATA})){
		# retrieve default wrap length
#		my $ThisWrap=$::Text::Wrap::columns;
		my $Size=$::FontHeight;
		my $Font='';

		if ($Command eq 'HEADING3'){$Font='HEADING3';}
		elsif ($Command eq 'HEADING2'){$Font='HEADING2'}
		elsif ($Command eq 'HEADING1'){$Font='HEADING1'};
		
		# # Second argument is wrap length
		# if ($#ArgList==2 and 
		#     ($ArgList[2]->QueryGeneralInfo->{FLOAT}||
		#      $ArgList[2]->QueryGeneralInfo->{INT})){
		#     $ThisWrap=$ArgList[2]->QueryGeneralInfo->{ATOMDATA};
		#     if ($ThisWrap<0){$ThisWrap=0;}
		# };

		my $Textstring=$ArgList[1]->QueryGeneralInfo->{ATOMDATA};
		if (defined $Textstring){
		    $Textstring=~s/^\"//; # remove leading and
		    $Textstring=~s/\"$//; # trailing double quotation mark
		    return &NewFromString(STRING => $Textstring,
					  PARENTOPERATOR=>'DISPLAY',
					  POSINPARENT=>0,
					  LISPLINK=>$LispLink,
					  STYLE => $Style,
					  WRAP => 1,
					  FONT => $Font);
		};
		# otherwise continue with normal function display
	    } elsif ($Command eq 'SEL' && ($#ArgList == 1 || $#ArgList == 2)){
		# The PROTECT command has two arguments:
		# - Expression
		# - Name of protection group
		# We are only drawing the Expression! 
		
		my $ProtStyle='PROTECTED';
		if ($#ArgList == 2){
		    my $GroupText=$ArgList[2]->QueryGeneralInfo->{'ATOMDATA'};
		    
		    # try to figure out the colour
		    if ($GroupText){
			$GroupText=~s/\"//g; # replace "" marks
			my $Ending; foreach $Ending('R','G','B','Y','P'){
			    if ($GroupText=~/
				_
				$Ending
				$
				/x){
				$ProtStyle='PROTECTED_'.$Ending;
			    };
			};
		    };
		};
		# We have to create a dummy list in order to be able to select the PROTECT.
		my @List=(&NewSeparator(' ',$LispLink));
		push @List,GO::New('LISPOBJECT'=>$ArgList[1],
				   'PARENTOPERATOR'=>'FUNCTION',
				   'LISPLINK'=>$LispLink
				   );
		push @List,&NewSeparator(' ',$LispLink);
		my $GO=&NewGOList('TYPE' => 'XLIST','COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@List,'STYLE' => $ProtStyle);
		return $GO;
	    } elsif ($Command eq 'EPS' && $#ArgList == 2 && $UE){
		my $Filename=$ArgList[1]->QueryGeneralInfo->{ATOMDATA};
		$Filename=~s/^\"//g; # remove leading "
		$Filename=~s/\"$//g; # remove trailing "
		my $Scale=$ArgList[2]->QueryGeneralInfo->{ATOMDATA};
		my $NewGO;
		if ($Filename){
		    $NewGO=&NewPicture(EPSFILE => $Filename, SCALE => $Scale,LISPLINK => $LispLink);
		};
		if ($NewGO){return $NewGO;};
		
		# Error handling
		return &NewFromString('STRING' => "Unknown EPS-image: $Filename",
				      'PARENTOPERATOR'=>$ParOp,
				      'POSINPARENT'=>$PosInParent,
				      'LISPLINK'=>$LispLink,
				      'STYLE' => $Style,
				      'WRAP' => $Wrap);
	    } elsif ($Command eq 'SYMAXX_GNUPLOT_DATA' && $UE && $::UIManager->QueryFeature(FEATURE => 'GNUPLOT')){
		return &NewFromString('STRING' => "#PLOT#",
				      'PARENTOPERATOR'=>$ParOp,
				      'POSINPARENT'=>$PosInParent,
				      'LISPLINK'=>$LispLink,
				      'STYLE' => $Style,
				      'WRAP' => $Wrap);
	    } 

	    elsif ($Command eq 'MCOND'){ #IF-statement	
		my @GOList;
		push @GOList,&NewSeparator('IF ',$LispLink);
		push @GOList,GO::New('LISPOBJECT' => $ArgList[1],
				   'PARENTOPERATOR' => $Command,
				   'POSINPARENT' => 1,
				   'LISPLINK'=>$LispLink);
		push @GOList,&NewSeparator(' THEN ',$LispLink);
		push @GOList,GO::New('LISPOBJECT' => $ArgList[2],
				   'PARENTOPERATOR' => $Command,
				   'POSINPARENT' => 2,
				   'LISPLINK'=>$LispLink);
		die $ArgList[3]->QueryAtomData unless $ArgList[3]->QueryAtomData eq '$TRUE' || $ArgList[3]->QueryAtomData eq 'T'; # Entry 3 is always TRUE
		if (!$ArgList[4]->QueryAtomData || $ArgList[4]->QueryAtomData ne '$FALSE'){
		    push @GOList,&NewSeparator(' ELSE ',$LispLink);
		    push @GOList,GO::New('LISPOBJECT' => $ArgList[4],
					 'PARENTOPERATOR' => $Command,
					 'POSINPARENT' => 3,
					 'LISPLINK'=>$LispLink);
		};		
		my $List=&NewGOList('TYPE' => 'XLIST','COMMAND' => $Command,'LISPLINK'=>$LispLink,'GOLIST'=>\@GOList);
		return $List;

	    } else { # any other function
		# Don't care here, what sort (user, system,...) it is.
		my @OuterGOList=(); # FUN Brackets(args)
		my @InnerGOList=(); # arg1,arg2,arg3
		#function name
		push @OuterGOList,GO::NewFromString('STRING' => $Command,
						    'PARENTOPERATOR'=>'FUNCTION',
						    'POSINPARENT'=>0,
						    'LISPLINK'=>$LispLink); # lisplink?
		
		my $Count;for ($Count=1;$Count<=$#ArgList;$Count++){
		    my $Elem=$ArgList[$Count] or die;
		    my $Separator='';
		    die "-$Command-$Elem-".join(" ",@ArgList) unless $Elem=~/LispList/;
		    if ($Count>1){$Separator=',';}
		    if ($Separator){push @InnerGOList,&NewSeparator($Separator,$LispLink);};
		    my $NewObj=GO::New('LISPOBJECT' => $Elem,
				       'PARENTOPERATOR' => 'FUNCTION',
				       'POSINPARENT' => $Count,
				       'LISPLINK'=>$LispLink);
		    push @InnerGOList,$NewObj;
		}; # foreach $Elem
		
		my $InnerXList=&NewGOList('TYPE' => 'XLIST','COMMAND' => 'PARTOF','LISPLINK'=>$LispLink,'GOLIST'=>\@InnerGOList);
		my $BracketType='ROUNDBRACKETS';
		if ($LOInfo->{'COMMANDARRAY'}){$BracketType='SQUAREBRACKETS';};
		my $BracketObject=&NewInBrackets('GRAPHOBJECT'=>$InnerXList,'LISPLINK'=>$LispLink,'BRACKETS'=> $BracketType);
		push @OuterGOList,$BracketObject;
		return &NewGOList('TYPE' => 'XLIST','COMMAND' => 'PARTOF', 'LISPLINK'=>$LispLink,'GOLIST'=>\@OuterGOList,'STYLE' => $Style);
	    }; # any other function
	}; # if not yet done
	$self->{'TYPE'}=$Command; # type means what kind of drawing is done, ...
	$self->{'COMMAND'}=$Command;  # command is used to determine, if something needs brackets.
	$self->{'LISPLINK'}=$LispLink;
	$self->{'POSINPARENT'}=$PosInParent;
	$self->{'PARENTOPERATOR'}=$ParOp;
	$self->{'STYLE'}=$Style;
	
	# To reduce overhead, we turn this only now into an object.
	# There are so many different other ways out of this sub.
	bless $self;    $::GOCount++; # Memory debug code
	
	if ($self->INeedBrackets('LISPOBJECT' => $Lispobject,'PARENTOPERATOR' => $ParOp,
				 'POSINPARENT' => $PosInParent)){
	    delete $self->{'STYLE'}; # brackets will do the background
	    return GO::NewInBrackets('GRAPHOBJECT'=>$self,'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS','STYLE' => $Style);
	}; 
	
	return $self;
    } else { # if command / not command
	# no? Then LispList.
	# This is a catch-all for any situation that we haven't thought of (for example rational expressions have to be handled this way)
	my @GOList=();
	my @ArgList=@{$Lispobject->QueryArgs} or die;
	my $Count;for ($Count=0;$Count<=$#ArgList;$Count++){
	    my $Elem=$ArgList[$Count] or die;
	    my $Separator='';
	    if ($Count>1){$Separator=',';}
	    if ($Separator){push @GOList,&NewSeparator($Separator,$LispLink);};
	    my $NewObj=GO::New('LISPOBJECT' => $Elem,
			       'PARENTOPERATOR' => 'LISPLIST',
			       'POSINPARENT' => $Count,
			       'LISPLINK'=>$LispLink);
	    push @GOList,&NewInBrackets('GRAPHOBJECT'=>$NewObj,'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS');
	}; # foreach $Elem
	my $XList=&NewGOList('TYPE' => 'XLIST','COMMAND' => 'LISPLIST','LISPLINK'=>$LispLink,'GOLIST'=>\@GOList);
	return &NewInBrackets('GRAPHOBJECT'=>$XList,'LISPLINK'=>$LispLink,'BRACKETS'=> 'ROUNDBRACKETS','STYLE' => $Style);
    };
};

sub Draw{
    my $self=shift or die;
    my %P=@_;
    my $Mode=$P{'MODE'}; die unless defined $Mode;
    
    my ($X,$Y);
    $X=$P{'X'}; die unless defined $X;
    $Y=$P{'Y'}; die unless defined $Y;
    
    if ($Mode eq 'NULLDRAW'){
	# Nulldraw means: query the geometry around $X, $Y. We may have it available
	# around (0,0).
	if ($self->{'LASTNULLGEOM'}){
	    my %BackGeom=%{$self->{'LASTNULLGEOM'}};
	    $BackGeom{'XMIN'}+=$X;
	    $BackGeom{'XMAX'}+=$X;
	    $BackGeom{'YMIN'}+=$Y;
	    $BackGeom{'YMAX'}+=$Y;
	    return \%BackGeom;
	};
    };
    my $Tag=$P{'TAG'}; die unless (defined $Tag || $Mode eq 'NULLDRAW');
    
    my $Geom={
	'XMIN' => $X,'XMAX' => $X,
	'YMIN'=>$Y,'YMAX'=>$Y};
    # In case of NULLDRAW, there is no absolute position information, but everything starts at 0.
    if ($Mode eq 'NULLDRAW'){$Geom->{'RELATIVE'}=1;}
    my $LastDraw;
    
    
    # do background UI
    if ($Mode eq 'DRAW'){
	my $Style=$self->{'STYLE'};
	if ($Style && $Style ne 'UNIT'){
	    my $MyGeom=$self->Draw('X'=>0,'Y'=>0,'MODE'=>'NULLDRAW');
	    my @Tags=[$Tag];
	    # is there a background colour setting?
	    
	    my $red=40000; my $green=40000; my $blue=40000;
	    my $Outline='';
	    if ($Style eq 'DARKRED'){$red=65535;$green=0;$blue=0;$Outline='black';} 
#	    elsif ($Style eq 'UNITS'){$red=65535;$green=65535;$blue=0;} 
	    elsif ($Style eq 'RED'){$red=65535;} # 
	    elsif ($Style eq 'GREEN'){$green=65535;}
	    elsif ($Style eq 'GREENBOX'){$red=0;$green=65535;$blue=0;$Outline='black';}
	    elsif ($Style eq 'PROTECTED'){$red=57000;$green=57000;$blue=57000;$Outline='orange';}
	    elsif ($Style eq 'PROTECTED_R'){$red=57000;$green=57000;$blue=57000;$Outline='red';}
	    elsif ($Style eq 'PROTECTED_G'){$red=57000;$green=57000;$blue=57000;$Outline='green';}
	    elsif ($Style eq 'PROTECTED_B'){$red=57000;$green=57000;$blue=57000;$Outline='blue';}
	    elsif ($Style eq 'PROTECTED_Y'){$red=57000;$green=57000;$blue=57000;$Outline='yellow';}
	    elsif ($Style eq 'PROTECTED_P'){$red=57000;$green=57000;$blue=57000;$Outline='magenta';}
	    elsif ($Style eq 'BLUE'){$blue=65535;}
	    elsif ($Style eq 'LIGHTGRAY'){$red=55000;$green=55000;$blue=65535;}
	    elsif ($Style eq 'LIGHTBLUE'){$blue=65535;}
	    elsif ($Style eq 'BOX'){$red=65535;$green=65535;$blue=65535;$Outline='black';}
	    elsif ($Style eq 'COMMAND'){		
		push @Tags,'THINGBACKGROUND';
		$red=60000;$green=60000;$blue=60000;
	    } elsif ($Style eq 'RESULT'){
		push @Tags,'THINGBACKGROUND';
		$red=55000;$green=55000;$blue=55000;
	    };
	    my $FillColor= sprintf('#%04x%04x%04x', $red, $green, $blue);
	    if (!$Outline){$Outline=$FillColor;};
	    $::GlobalCanvas->createRectangle($X+$MyGeom->{'XMIN'},$Y+$MyGeom->{'YMIN'},
					     $X+$MyGeom->{'XMAX'},$Y+$MyGeom->{'YMAX'},
					     -fill => $FillColor,-tags => [$Tag,'THINGBACKGROUND'],-outline => $Outline);
	}; #if background
    }; # if draw mode
    if ($self->{'TYPE'} eq 'ATOM'){
	my $AtomText=$self->{DATA1};
	my $Color='black';

#	if ($AtomText eq 'EXPAND'){
#	    print join(":",%{$self->{LISPLINK}->{GENERALINFO}});
#	};

	# Highlighting of known keywords and units
	if ($::GlobalFontLock){
	    if ($self->{'STYLE'} && 
		$self->{'STYLE'} eq 'UNIT'){
		$Color='blue';
	    } elsif ($::MaximaSymbols->{$AtomText}){
		$Color='darkgreen';
	    };

	};
	
	my $Fontname=$self->{FONT};
	if (!$self->{'WRAP'}){
	    # Code duplicated to optimize performance
	    if (!$Fontname){
		$LastDraw=&DrawSimpleText($X,$Y,$AtomText,$Mode,$Tag,$Color);
	    } else {
		$LastDraw=&DrawSimpleTextAnyFont($X,$Y,$AtomText,$Fontname,$Mode,$Tag,$Color);
	    };
	    &MergeGeom($Geom,$LastDraw);
	} else {
	    $AtomText=~s/\\n/\n/g; # replace each backslash-n with a newline

	    # setup default wrap length
	    $Text::Wrap::columns=$::WrapDefaultColumns;
	    if ($Fontname){
		$Text::Wrap::columns=$::FontWrapColumns_{$Fontname};		
	    };

	    $AtomText=wrap('','',$AtomText);
	    my @Textlist=split(/\n/,$AtomText);
	    my $DrawY=$Y;
	    my $Line; foreach $Line(@Textlist){
		if (!$self->{FONT}){
		    $LastDraw=&DrawSimpleText($X,$DrawY,$Line,$Mode,$Tag,$Color);
		} else {
		    $LastDraw=&DrawSimpleTextAnyFont($X,$DrawY,$Line,$self->{FONT},$Mode,$Tag,$Color);
		};
		$DrawY+=&Height($LastDraw);
		&MergeGeom($Geom,$LastDraw);
	    }; # foreach line
	};
    } elsif ($self->{'TYPE'} eq 'SEPARATOR'){
	$LastDraw=&DrawSimpleText($X,$Y,$self->{'ARG1'},$Mode,$Tag);
	&MergeGeom($Geom,$LastDraw);
    } elsif ($self->{'TYPE'} eq 'MEXPT'){ 
	my $LowerGeom=$self->{'ARGS'}->[0]->Draw('X' => 0,'Y' => 0,'MODE' => 'NULLDRAW');
	my $UpperGeom=$self->{'ARGS'}->[1]->Draw('X' => 0,'Y' => 0,'MODE' => 'NULLDRAW');
	my $AddHeight=&UpperHeight($LowerGeom)+&LowerHeight($UpperGeom);
	$LastDraw=$self->{'ARGS'}->[0]->Draw('X' => $Geom->{'XMAX'},'Y' => $Y,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LastDraw);
	$LastDraw=$self->{'ARGS'}->[1]->Draw('X' => $Geom->{'XMAX'},'Y' => $Y-$AddHeight,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LastDraw);
    } elsif ($self->{'TYPE'} eq 'XLIST') {
	# do arguments
	die unless $self->{'ARGS'};
	my @DataArray=@{$self->{'ARGS'}};
	my $Count;for($Count=0;$Count<=$#DataArray;$Count++){
	    # leave one pixel extra between any two elements
	    if ($Count !=0){$Geom->{'XMAX'}++;};
	    my $Elem=$self->{'ARGS'}->[$Count];
	    $LastDraw=$Elem->Draw('X' => $Geom->{'XMAX'},
				  'Y' => $Y,'MODE'=> $Mode,'TAG' => $Tag);
	    &MergeGeom($Geom,$LastDraw);
	};
    } elsif ($self->{'TYPE'} eq 'YLIST') {
	die unless $self->{'ARGS'};
	my @DataArray=@{$self->{'ARGS'}};
	my $ArgNumber=$#DataArray;
	
	# find out the height above the baseline
	my $CenterElem=($#DataArray)/2; # 1 elem: => 0 , 2 => 0.5, 3=> 1 etc  
	my $UpperHeight=0;
	my $Count;
	my @ElemUpperHeight;
	for ($Count=0;$Count<=$#DataArray;$Count++){
	    my $Arg=$DataArray[$Count];
	    my $ArgGeom=$Arg->Draw('MODE' => 'NULLDRAW','X'=>0,'Y'=>0);
	    # if the element is fully over the baseline, then it adds its full height
	    if ($Count<$CenterElem){
		$UpperHeight+=&Height($ArgGeom);
		# but if it is on the baseline, then it adds only half its height.
	    } elsif ($Count==$CenterElem){
		$UpperHeight+=&UpperHeight($ArgGeom);
	    };
	    $ElemUpperHeight[$Count]=&UpperHeight($ArgGeom);
	}; # for $Count
	
	# now we have the upper edge. 
	# The element has to be moved down by its upper height, because the draw commands takes the baseline as reference, not the upper edge.
	$Geom->{'YMIN'}-=$UpperHeight;
	$Geom->{'YMAX'}-=$UpperHeight;
	for($Count=0;$Count<=$#DataArray;$Count++){
	    my $Elem=$self->{'ARGS'}->[$Count];
	    # +1: leave some space
	    $LastDraw=$Elem->Draw('X' => $X,
				  'Y' => $Geom->{'YMAX'}+$ElemUpperHeight[$Count]+1,'MODE'=> $Mode,'TAG' => $Tag);
	    &MergeGeom($Geom,$LastDraw);
	};
    } elsif ($self->{'TYPE'} eq 'FRACTION'){
	my $UpperGeom=$self->{'ARGS'}->[0]->Draw('MODE' => 'NULLDRAW','X'=>0,'Y'=>0);		
	my $LowerGeom=$self->{'ARGS'}->[1]->Draw('MODE' => 'NULLDRAW','X'=>0,'Y'=>0);		
	my $MaxWidth=MySubs::Max(&Width($UpperGeom),&Width($LowerGeom));
	my $XExtra=$MaxWidth/10;
	
	# Y-distance between baseline and start left upper / lower corner
	
	my $YUpper=&LowerHeight($UpperGeom)+2;
	my $YLower=&UpperHeight($LowerGeom)+2;

	my $XShiftUpper=($MaxWidth-&Width($UpperGeom))/2+$XExtra/2;
	my $XShiftLower=($MaxWidth-&Width($LowerGeom))/2+$XExtra/2;
	$UpperGeom=$self->{'ARGS'}->[0]->Draw('X' => $X+$XShiftUpper,'Y' => $Y-$YUpper,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$UpperGeom);
	$LowerGeom=$self->{'ARGS'}->[1]->Draw('X' => $X+$XShiftLower,'Y' => $Y+$YLower,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LowerGeom);
	my $LineGeom=DrawSimpleLine('LINEGEOM' => [$X,$Y,$X+$MaxWidth+$XExtra,$Y],'TAG' => $Tag,'MODE' => $Mode);
	&MergeGeom($Geom,$LineGeom);
    } elsif (($self->{'TYPE'} eq 'ROUNDBRACKETS')||($self->{'TYPE'} eq 'SQUAREBRACKETS')||($self->{'TYPE'} eq 'VERTBARS')||($self->{'TYPE'} eq 'INTEGRATE')||($self->{'TYPE'} eq 'SUM')){
	my $BracketOpen; my $BracketClose; 
	if ($self->{'TYPE'} eq 'ROUNDBRACKETS'){
	    $BracketOpen='('; $BracketClose=')';
	} elsif ($self->{'TYPE'} eq 'SQUAREBRACKETS'){
	    $BracketOpen='['; $BracketClose=']';
	} elsif ($self->{'TYPE'} eq 'VERTBARS'){
	    $BracketOpen='|open'; $BracketClose='|close';
	} elsif ($self->{'TYPE'} eq 'INTEGRATE'){
	    $BracketOpen='INTEGRATE'; $BracketClose='';
	} elsif ($self->{'TYPE'} eq 'SUM'){
	    $BracketOpen='SUM'; $BracketClose='';
	};
	
	my $ArgGeom=$self->{'ARGS'}->[0]->Draw('X'=>0,'Y'=>0,'MODE' => 'NULLDRAW');
	$LastDraw=&DrawBracket('X' => $Geom->{'XMAX'},'Y' => $Y,'TYPE' => $BracketOpen,
			       'GEOM' => $ArgGeom,'MODE' => ,$Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LastDraw);
	$LastDraw=$self->{'ARGS'}->[0]->Draw('X'=>$Geom->{'XMAX'},'Y'=>$Y,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LastDraw);
	if ($BracketClose){
	    $LastDraw=&DrawBracket('X' => $Geom->{'XMAX'},'Y' => $Y,'TYPE' => $BracketClose,
				   'GEOM' => $ArgGeom,'MODE' => ,$Mode,'TAG' => $Tag);
	    &MergeGeom($Geom,$LastDraw);
	};
    } elsif ($self->{'TYPE'} eq 'SQRT'){
	my $ArgGeom=$self->{'ARGS'}->[0]->Draw('X'=>0,'Y'=>0,'MODE' => 'NULLDRAW');
	my @Pointlist=();
	my $LowerHeight=&LowerHeight($ArgGeom);
	my $UpperHeight=&UpperHeight($ArgGeom)+1; # leave one pixel space between the contents and the upper line
	my $DX1=$LowerHeight/4;
	my $DX2=$UpperHeight/4;
	my $DX0=$DX1+$DX2/2;
	push @Pointlist,$X;push @Pointlist,$Y;
	push @Pointlist,$X+$DX0;push @Pointlist,$Y;
	push @Pointlist,$X+$DX0+$DX1;push @Pointlist,$Y+$LowerHeight;
	push @Pointlist,$X+$DX0+$DX1+$DX2;push @Pointlist,$Y-$UpperHeight;
	push @Pointlist,$X+$DX0+$DX1+$DX2+&Width($ArgGeom)+3;push @Pointlist,$Y-$UpperHeight;
	push @Pointlist,$X+$DX0+$DX1+$DX2+&Width($ArgGeom)+3;push @Pointlist,$Y-$UpperHeight+$DX0;
	$LastDraw=DrawSimpleLine('LINEGEOM' => \@Pointlist,'TAG' => $Tag,'MODE' => $Mode);
	&MergeGeom($Geom,$LastDraw);
	$LastDraw=$self->{'ARGS'}->[0]->Draw('X'=>$X+$DX0+$DX1+$DX2+2,'Y'=>$Y,'MODE' => $Mode,'TAG' => $Tag);
	&MergeGeom($Geom,$LastDraw);
    } elsif ($self->{'TYPE'} eq 'AT'){
	$LastDraw=$self->{'ARGS'}->[0]->Draw('X'=>$X,'Y'=>$Y,'MODE' => $Mode,'TAG' => $Tag);	
	&MergeGeom($Geom,$LastDraw);
	my @Pointlist=($LastDraw->{'XMAX'},$LastDraw->{'YMIN'},$LastDraw->{'XMAX'},$LastDraw->{'YMAX'}); 
	$LastDraw=DrawSimpleLine('LINEGEOM' => \@Pointlist,'TAG' => $Tag,'MODE' => $Mode);
	&MergeGeom($Geom,$LastDraw);
	$LastDraw=$self->{'ARGS'}->[1]->Draw('X'=>$LastDraw->{'XMAX'},'Y'=>$LastDraw->{'YMAX'},'MODE' => $Mode,'TAG' => $Tag);	
	&MergeGeom($Geom,$LastDraw);  
    }  elsif ($self->{'TYPE'} eq 'PICTURE'){
	my $EPSFile=$self->{'EPSFILE'};
	my $Handle=$self->{HANDLE} or die;
	my $Photo;
	$Photo=$::UIManager->GetPicture(HANDLE => $Handle,EPSFILE => $EPSFile);
	if ($Photo){
	    if ($Mode eq 'DRAW'){
		if (!$::PrintMode){ # this is the normal case
		    # the picturename is reused as handle, so that we can delete it later, if needed.
		    $::GlobalCanvas->createImage($X,$Y+1,-image => $Handle,-anchor => "nw",-tags => [$Tag, $Handle,'PICTURE']) or die;
		} else {
		    # We are preparing a printout. 
		    # Now pictures are not drawn (they are replaced later throught their postscript files). 
		    # Instead, we put some unique text in their place. It gets replaced later  in the PS file. Messy, but works.
		    $::GlobalCanvas->createText($X,$Y+1,-anchor => 'sw', -text => "NobodyRightInHisMindWouldUseThisString\_$Handle\_Yes",-tags=>[$Tag,$Handle]);
		}; # if PrintMode	
		$self->{'EPSX'}=$X;
		$self->{'EPSY'}=$Y+1;
		
	    }; # if draw
	    $Geom->{'XMAX'}+=$Photo->width;
	    $Geom->{'YMAX'}+=$Photo->height+1;
	};
    } else {die 'undefined type'.$self->{'TYPE'};}
    my %GeomCopy=%{$Geom};
    if ($Mode eq 'DRAW'){
	$self->{'LASTDRAWGEOM'}=\%GeomCopy;
	# please note: We can't simply delete LastNullGeom at this point,
	# it will be needed later.
    } else {
	# transform into origin (X=0,Y=0)
	$GeomCopy{'XMIN'}-=$X;
	$GeomCopy{'XMAX'}-=$X;
	$GeomCopy{'YMIN'}-=$Y;
	$GeomCopy{'YMAX'}-=$Y;
	$self->{'LASTNULLGEOM'}=\%GeomCopy;
    };
    return $Geom;
};
sub MergeGeom{
    #An element is appended to the right end of a row
    my $Geom1=shift or die;
    my $Geom2=shift or die;	
    if ($Geom1->{'RELATIVE'} && $Geom2->{'RELATIVE'}){
    } elsif (!$Geom1->{'RELATIVE'} && !$Geom2->{'RELATIVE'}){} else {die "Mixed relative / absolute mode!";};
    $Geom1->{'XMIN'}=MySubs::Min($Geom1->{'XMIN'},$Geom2->{'XMIN'});
    $Geom1->{'XMAX'}=MySubs::Max($Geom1->{'XMAX'},$Geom2->{'XMAX'});
    $Geom1->{'YMIN'}=MySubs::Min($Geom1->{'YMIN'},$Geom2->{'YMIN'});
    $Geom1->{'YMAX'}=MySubs::Max($Geom1->{'YMAX'},$Geom2->{'YMAX'});
};	
sub DrawSimpleText{
    my $X=shift; die unless defined $X;
    my $Y=shift; die unless defined $Y;
    my $Text=shift; die unless defined $Text;
    my $Mode=shift or die;
    my $Tag=shift; die unless (defined $Tag || $Mode eq 'NULLDRAW');
    my $Color=shift; if (!defined $Color){$Color='black';};
   $::GlobalCanvas->createText($X,$Y,-text =>$Text,-anchor =>'w',-font => 'FRAGFONT',-tags => $Tag,-fill => $Color) unless ($Mode ne 'DRAW');
    my $LengthInPixels=&Stringlength($Text);
    my $Geom={XMIN => $X,
	      XMAX=>$X+$LengthInPixels,
	      YMIN=>$Y-$::UpperFontHeight,
	      YMAX=>$Y+$::LowerFontHeight};
    if ($Mode eq 'NULLDRAW'){$Geom->{RELATIVE}=1;};
    return $Geom;
};
sub DrawSimpleTextAnyFont{
    my $X=shift; die unless defined $X;
    my $Y=shift; die unless defined $Y;
    my $Text=shift; die unless defined $Text;
    my $Fontname=shift or die;
    my $Mode=shift or die;
    my $Tag=shift; die unless (defined $Tag || $Mode eq 'NULLDRAW');
   $::GlobalCanvas->createText($X,$Y,-text =>$Text,-anchor =>'w',-font => $Fontname,-tags => $Tag) unless ($Mode ne 'DRAW');
    my $LengthInPixels=&StringlengthAnyFont($Text,$Fontname);
    my $Geom={XMIN => $X,
	      XMAX=>$X+$LengthInPixels,
	      YMIN=>$Y-$::UpperFontHeight_{$Fontname},
	      YMAX=>$Y+$::LowerFontHeight_{$Fontname}};
    if ($Mode eq 'NULLDRAW'){$Geom->{RELATIVE}=1;};
    return $Geom;
};
sub DrawSimpleLine{
    my %P=@_;
    die unless $P{'LINEGEOM'};
    my @Linegeom=@{$P{'LINEGEOM'}};
    my $Mode=$P{'MODE'};
#    if ($Mode eq 'DRAW'){
#	print join(":",@Linegeom)."\n";};
    my $Tag=$P{'TAG'}; die unless (defined $Tag || $Mode eq 'NULLDRAW');
    if ($Mode eq 'DRAW'){
	$::GlobalCanvas->createLine(@Linegeom,-tags => $Tag);   
    }; # if mode is DRAW
    my $XMin=shift @Linegeom; my $YMin=shift @Linegeom;
    my $XMax=$XMin; my $YMax=$YMin;
    my $XNew; my $YNew; while (defined($XNew=shift @Linegeom)){
	$YNew=shift @Linegeom;
	$XMin=MySubs::Min($XMin,$XNew-1);
	$XMax=MySubs::Max($XMax,$XNew+1);
	$YMin=MySubs::Min($YMin,$YNew-1);
	$YMax=MySubs::Max($YMax,$YNew+1);	
    };
    my $Geom={'XMIN' => $XMin,
	      'XMAX'=> $XMax,
	      'YMIN'=> $YMin,
	      'YMAX'=> $YMax};
    if ($Mode eq 'NULLDRAW'){$Geom->{'RELATIVE'}=1;};
    return $Geom;
};
sub QueryLeadingMinus{
    my $self=shift or die;
    if (!defined $self->{'ARG1'}){return 0;};
    return $self->{'ARG1'}=~/^\-/; # look for leading minus sign in text
};
sub Stringlength{
    ######################################################################################
    # Returns the graphical length of the string in pixels. Since the Tk-function is     #
    # really slow (fractions of second), we do it ourselves using a hash of character    #
    # widths. If the width for the current character has not been calculated yet, do so. #
    ######################################################################################
    my $String=shift;
    my @Array=split(//,$String); # split up into characters
    if (!defined $::StringWidth{$String}){
	my $Length=0;
	my $Char;
	while ($#Array>=0){
	    $Char=shift @Array;
	    if (!defined($::CharacterWidth{$Char})){
		$::CharacterWidth{$Char}=$::GlobalCanvas->fontMeasure('FRAGFONT',$Char);
	    };
	    $Length+=$::CharacterWidth{$Char};    
	};
	$::StringWidth{$String}=$Length;
	return $Length;
    };
    return $::StringWidth{$String};
};

# the 'AnyFont' subroutines may be unused.
sub StringlengthAnyFont{
    ######################################################################################
    # Returns the graphical length of the string in pixels.                              #
    ######################################################################################
    my $String=shift;
    my $Font=shift;
    my @Array=split(//,$String); # split up into characters
    if (!defined $::StringWidthAnyFont{$Font} || !defined $::StringWidthAnyFont{$Font}->{$String}){
	my $Length=0;
	my $Char;
	while ($#Array>=0){
	    $Char=shift @Array;
	    if (
		!defined($::CharacterWidthAnyFont{$Font}) 
		|| !defined($::CharacterWidthAnyFont{$Font}->{$Char})){
		$::CharacterWidthAnyFont{$Font}->{$Char}=$::GlobalCanvas->fontMeasure($Font,$Char);
	    };
	    $Length+=$::CharacterWidthAnyFont{$Font}->{$Char};    
	};
	$::StringWidthAnyFont{$Font}->{$String}=$Length;
	return $Length;
    };
    return $::StringWidthAnyFont{$Font}->{$String};
};
sub Height{
    my $Geom=shift or die;
    return $Geom->{'YMAX'}-$Geom->{'YMIN'};
};
sub Width{
    my $Geom=shift or die;
    return $Geom->{'XMAX'}-$Geom->{'XMIN'};
};
sub LowerHeight{
    my $Geom=shift or die;
    die unless $Geom->{'RELATIVE'};
    return $Geom->{'YMAX'};
};
sub UpperHeight{
    my $Geom=shift or die;
    die unless $Geom->{'RELATIVE'};
    return -$Geom->{'YMIN'};
};
sub QueryCoordinatesWithin{
    my $self=shift or die;
    my %P=@_;  
    my $X=$P{'X'}; die unless defined $X;
    my $Y=$P{'Y'}; die unless defined $Y;
    my $LastDrawGeom=$self->{'LASTDRAWGEOM'} or die;
    if ($X>=$LastDrawGeom->{'XMIN'} && 
	$X<=$LastDrawGeom->{'XMAX'} && 
	$Y>=$LastDrawGeom->{'YMIN'} && 
	$Y<=$LastDrawGeom->{'YMAX'}){
	return 1;
    };
    return 0;
};
sub QueryGraphObjectFromCoordinates{
    my $self=shift or die;
    my %P=@_;  
    my $X=$P{'X'}; die unless defined $X;
    my $Y=$P{'Y'}; die unless defined $Y;
    my $LastDrawGeom=$self->{'LASTDRAWGEOM'};
    
    # if it has not been drawn yet, we can't have clicked on it.
    return undef unless defined $LastDrawGeom;
    
    # Are the coordinates within this object?
    if ($X>=$LastDrawGeom->{'XMIN'} && 
	$X<=$LastDrawGeom->{'XMAX'} && 
	$Y>=$LastDrawGeom->{'YMIN'} && 
	$Y<=$LastDrawGeom->{'YMAX'}){
	# Then try to find a smaller object still.
	my $Width=&Width($LastDrawGeom);
	my $Height=&Height($LastDrawGeom);

	if ($self->{'ARGS'}){
	    my @Args=@{$self->{'ARGS'}};
	    my $Count; for ($Count=0;$Count<=$#Args; $Count++){
		my $Arg=$Args[$Count];
		my $SmallerObjectData=$Arg->QueryGraphObjectFromCoordinates('X'=>$X,'Y'=>$Y);
		# We don't return a smaller object, if the size is the same. Otherwise, there would be no way to select this object.
		# Oops... This does not work. Disabled (has to work with SIN(x), too)
		if ($SmallerObjectData && (1 || &Width($SmallerObjectData->{'GEOM'})<$Width || &Height($SmallerObjectData->{'GEOM'}) <$Height))			{
		    $SmallerObjectData->{'DEPTH'}++;
		    return $SmallerObjectData;
		};
	    }; # for Count
	}; # if this has args
	return {'OBJECT' => $self,'DEPTH' => 0,'GEOM' => $LastDrawGeom};
    }; # if within this object
    return undef;
};
sub QueryGOFromLispLink{
    my $self=shift or die;
    my %P=@_;  
    my $LispLink=$P{'LISPLINK'} or die;    
    if ($self->{'LISPLINK'} eq $LispLink){return $self;};
    if ($self->{'ARGS'}){
	my @Args=@{$self->{'ARGS'}};
	my $Count; for ($Count=0;$Count<=$#Args; $Count++){
	    my $Arg=$Args[$Count];
	    my $GO=$Arg->QueryGOFromLispLink(@_);
	    if ($GO){return $GO;};
	}; # for Count
    }; # if args
    return undef;
};
sub QueryLispLink{
    my $self=shift or die;
    die unless defined $self->{'LISPLINK'};
    return $self->{'LISPLINK'};
};
sub QueryGeom{
    my $self=shift or die;
    if (!$self->{'LASTNULLGEOM'}){
	# if the geometry info does not yet exist, create it.
	$self->Draw('X'=>0,'Y'=>0,'MODE' => 'NULLDRAW');
    };
    return $self->{'LASTNULLGEOM'};
};	
sub QueryDrawGeom{
    my $self=shift or die;
    die unless $self->{'LASTDRAWGEOM'};
    return $self->{'LASTDRAWGEOM'};
};	
sub DrawBracket{
    my %P=@_;
    my $X=$P{'X'}; die unless defined $X;
    my $Y=$P{'Y'}; die unless defined $Y;
    my $Geom=$P{'GEOM'} or die;
    my $Type=$P{'TYPE'} or die;
    my $Mode=$P{'MODE'} or die;
    my $Tag=$P{'TAG'}; die unless (defined $Tag || $Mode eq 'NULLDRAW');
    
    # empty brackets have a nonzero-size - for example []
    if (&Height($Geom)<$::FontHeight){
	$Geom->{'YMIN'}=MySubs::Min($Geom->{'YMIN'},-$::UpperFontHeight);
	$Geom->{'YMAX'}=MySubs::Max($Geom->{'YMAX'},$::UpperFontHeight);
    };
    my $Upper=$Y-&UpperHeight($Geom);
    my $Lower=$Y+&LowerHeight($Geom);
    my $Height=&Height($Geom);
    # if the integral sign contains only one line of text, make it bigger.
    # Still: Symaxx 2 proudly features the most ugly int-sign on the planet :-)
    if ($Type eq 'INTEGRATE' && $Height<2*$::FontHeight){
	my $Heightdiff=2*$::FontHeight-$Height;
	$Upper-=$Heightdiff/2;$Lower+=$Heightdiff/2;
	$Height=$Lower-$Upper;
    };
    my $a=int($Height/6);
    $a=MySubs::Max($a,3);
    $a=MySubs::Min($a,8);    
    my (@X1,@Y1,@X2,@Y2);
#    print "$a $Upper $Lower\n";

    my @Pointlist=();
    if ($Type eq '('){
	push @Pointlist,$X+$a;
	push @Pointlist,$Upper;
	push @Pointlist,$X;
	push @Pointlist,$Upper+$a;
	push @Pointlist,$X;
	push @Pointlist,$Lower-$a;
	push @Pointlist,$X+$a;
	push @Pointlist,$Lower;
    } elsif ($Type eq ')'){
	push @Pointlist,$X;
	push @Pointlist,$Upper;
	push @Pointlist,$X+$a;
	push @Pointlist,$Upper+$a;
	push @Pointlist,$X+$a;
	push @Pointlist,$Lower-$a;
	push @Pointlist,$X;
	push @Pointlist,$Lower;
    }  elsif ($Type eq '['){
	push @Pointlist,$X+$a;
	push @Pointlist,$Upper;
	push @Pointlist,$X;
	push @Pointlist,$Upper;
	push @Pointlist,$X;
	push @Pointlist,$Lower;
	push @Pointlist,$X+$a;
	push @Pointlist,$Lower;
    } elsif ($Type eq ']'){
	push @Pointlist,$X;
	push @Pointlist,$Upper;
	push @Pointlist,$X+$a;
	push @Pointlist,$Upper;
	push @Pointlist,$X+$a;
	push @Pointlist,$Lower;
	push @Pointlist,$X;
	push @Pointlist,$Lower;
    } elsif ($Type eq '|'){
	push @Pointlist,$X;
	push @Pointlist,$Upper;
	push @Pointlist,$X;
	push @Pointlist,$Lower;
    } elsif ($Type eq '|open'){
	push @Pointlist,$X;
	push @Pointlist,$Upper;
	push @Pointlist,$X;
	push @Pointlist,$Lower;
    } elsif ($Type eq '|close'){
	push @Pointlist,$X+1;
	push @Pointlist,$Upper;
	push @Pointlist,$X+1;
	push @Pointlist,$Lower;
    } elsif ($Type eq 'INTEGRATE'){
	my $a=MySubs::Max($a,4); # somewhat bigger than the default
	push @Pointlist,$X;push @Pointlist,$Lower-$a;
	push @Pointlist,$X+$a/2;push @Pointlist,$Lower;
	push @Pointlist,$X+$a;push @Pointlist,$Lower-$a;
	push @Pointlist,$X+$a;push @Pointlist,$Upper+$a;
	push @Pointlist,$X+3*$a/2;push @Pointlist,$Upper;
	push @Pointlist,$X+2*$a;push @Pointlist,$Upper+$a;
    } elsif ($Type eq 'SUM'){
	my $Middle=($Upper+$Lower)/2;
	push @Pointlist,$X+2*$a;push @Pointlist,$Lower-$a/2;
	push @Pointlist,$X+2*$a;push @Pointlist,$Lower;
	push @Pointlist,$X;push @Pointlist,$Lower;
	push @Pointlist,$X+2*$a;push @Pointlist,$Middle;
	push @Pointlist,$X;push @Pointlist,$Upper;
	push @Pointlist,$X+2*$a;push @Pointlist,$Upper;
	push @Pointlist,$X+2*$a;push @Pointlist,$Upper+$a;
     } else {
	die $Type;
    };
    my $RetGeom=DrawSimpleLine('LINEGEOM' => \@Pointlist,'TAG' => $Tag,'MODE' => $Mode);
    if ($Type eq ')'){$RetGeom->{'XMAX'}+=2;}
    elsif ($Type eq 'SUM' || $Type eq '|open' || $Type eq '|close' || $Type eq '|'){
	$RetGeom->{'XMAX'}+=2;};
    return $RetGeom;
};
sub HighlightPreciseSelection{
    my $self=shift or die;
    my %P=@_;
    my $WantedLisplink=$P{'LISPLINK'}; die unless $WantedLisplink || $P{'FORCE'};
    my $Mode=$P{'MODE'} or die;

    if ($P{'FORCE'} || $self->{'LISPLINK'} eq $WantedLisplink){
	# this is the most precise selection.
	# Draw a cursor
	my $LastDrawGeom=$self->{'LASTDRAWGEOM'} or die;
	if ($Mode eq 'CURSOR'){
	    $::GlobalCanvas->delete('PRECISESELECTIONCURSOR');
	    $::GlobalCanvas->createRectangle($LastDrawGeom->{'XMIN'},
					     $LastDrawGeom->{'YMIN'},
					     $LastDrawGeom->{'XMAX'},
					     $LastDrawGeom->{'YMAX'},
					     -tags => 'PRECISESELECTIONCURSOR');
	}
    } else {
	if ($self->{'ARGS'}){
	    my $Arg; foreach $Arg(@{$self->{'ARGS'}}){
		if ($Arg->HighlightPreciseSelection(%P)){
		    return 1;
		};
	    };
	};
    };
}
sub DeleteNullgeom{
    my $self=shift or die;
    my %P=@_;
    if ($self->{'ARGS'}){
	my $Arg; foreach $Arg(@{$self->{'ARGS'}}){
	    $Arg->DeleteNullgeom;
	};
    };
    if (!$P{'LOWERLEVELONLY'}){
	# This is somewhat dangerous: If NullGeom IS used at a later stage,
	# it has to be recalculated. Then it doesn't save memory,
	# but causes massive computational overhead.
	# But I am pretty confident that it is only needed for the 
	# top level GO (when moving).
	# Anyway, the memory saved is almost trivial.

#	$self->{'NULLGEOM'}='kaboom'; # test if it is needed later
	delete $self->{'LASTNULLGEOM'};
    };
};
sub DESTROY{  
    # Memory debug code
    $::GOCount--;
}
1;


