package Palette;

# Contents: Palette window (allows to choose selected commands by mouse)

use Tk;
use strict;

use Tk::NoteBook;

sub New{
    my $self={};bless $self;
    my @L=(); # list of widgets
    # global variable - only one dialog allowed at a time.
    my $M=$::MainWin;
    if (&AreYouOK){$::Palette->destroy;};
    $::Palette=$M->Toplevel;
    my $D=$::Palette;
    $D->title("Palette");
   
    $::Palette->bind('<Escape>',sub{$::Palette->destroy;});
    my @AllButtonData=();
    
    # Keywords:
    # 'T' => Title
    # 'P' => Picture
    # 'C' => Command
    # EXPR: Will be replaced by selection
    # otherwise start all new variables with My (MyVar, MyStart, MyStop)
    push @AllButtonData,'Int/Diff';
    push @AllButtonData,
    [
     {'P' => 'Integrate.gif',
      'C' => 'INTEGRATE(EXPR,MyVar)'},
     {'P' => 'Integrate_a_b.gif',
      'C' => 'INTEGRATE(EXPR,MyVar,MyStart,MyStop)'},
     {'P' => 'Integrate_Noun.gif',
      'C' => "'INTEGRATE(EXPR,MyVar)"},
     {'P' => 'Diff.gif',
      'C' => 'DIFF(EXPR,MyVar)' }, 
     {'P' => 'DiffN.gif',
      'C' => 'DIFF(EXPR,MyVar,MyN)' }
     ]; # Int/Diff button
    push @AllButtonData,'Expand';
    push @AllButtonData,
    [
     {'P' => 'Expand.gif',
      'C' => 'EXPAND(EXPR)'},
     {'P' => 'Xthru.gif',
      'C' => 'XTHRU(EXPR)'},
     {'P' => 'Multthru.gif',
      'C' => 'MULTTHRU(EXPR)'}
     ]; 
    push @AllButtonData,'Simplify';
    push @AllButtonData,
    [
     {'T' => 'RADCAN',
      'C' => 'RADCAN(EXPR)'},
     {'T' => 'RATSIMP',
      'C' => 'RATSIMP(EXPR)'},
     {'T' => 'FULLRATSIMP',
      'C' => 'FULLRATSIMP(EXPR)'}
     ]; 
    push @AllButtonData,'Trig';
    push @AllButtonData,
    [
     {'P' => 'Trigsimp.gif',
      'C' => 'TRIGSIMP(EXPR)'},
     {'P' => 'Trigexpand.gif',
      'C' => 'TRIGEXPAND(EXPR)'},
     {'P' => 'Trigreduce.gif',
      'C' => 'TRIGREDUCE(EXPR)'}
     ]; 
    
    push @AllButtonData,'Text / Picture';
    push @AllButtonData,
    [
     {'T' => 'EPS picture normal scale',
      'C' => 'EPS("'.$::CodeDir.'/Bitmaps/Something.eps",1)'},
     {'T' => 'EPS picture half scale',
      'C' => 'EPS("'.$::CodeDir.'/Bitmaps/Something.eps",0.5)'},
     {'T' => 'EPS picture double scale',
      'C' => 'EPS("'.$::CodeDir.'/Bitmaps/Something.eps",2)'},
     {'T' => 'Plain text',
      'C' => 'TEXT("EXPR")'},
     {'T' => 'Big heading',
      'C' => 'HEADING1("EXPR")'},
     {'T' => 'Medium heading',
      'C' => 'HEADING2("EXPR")'},
     {'T' => 'Small heading',
      'C' => 'HEADING3("EXPR")'},
     {'T' => 'Heading, text,...',
      'C' => 'YLIST(HEADING1("Heading"),TEXT("EXPR"),HEADING2("Heading"),TEXT("Text"))'}
     ]; 
    if ($::MaximaUseNumerical){
	push @AllButtonData,'Numerical / Table';
	push @AllButtonData,
	[
	 {'P' => 'CreateTable.gif',
	  'C' => 'makelist([x,x^2],x,0,10)'},
     {'P' => 'F_of_table.gif',
      'C' => 'TableOpApply([[-30,a0],[-29,b0],[-28,c0]],Lambda([Indep,var1],var1+1))'},
	 {'P' => 'F_of_tables.gif',
	  'C' => 'TableOpApply([[-30,a],[-29,b],[-28,c]],[[-30,a0],[-29,b0],[-28,c0]],Lambda([Indep,var1,var2],var1/var2))'},
	 {'P' => 'Diff_Numerical.gif',
	  'C' => 'NumDiff(x^2+1,x,1)'},
	 {'P' => 'Integrate_Numerical.gif',
	  'C' => 'NumIntegrate(x^2,x,0,1)'},
	 {'P' => 'LinearInterpolation.gif',
	  'C' => 'f(x):=LinInterp(x,[[0,0],[1,1],[2,4],[3,9]])'},
	 ]; 
    };
    my @AllButtonData_=@AllButtonData;
    while ($#AllButtonData_ >=0){
	my $ButtonText=shift @AllButtonData_;
	my @ThisButtonData=@{shift @AllButtonData_};
	my @ThisButtonDataStructure=();
	my $NextLine;
	my %ThisButtonCommand;
	while ($NextLine=shift @ThisButtonData){
	    %ThisButtonCommand=%{$NextLine};
	    my $Command=$ThisButtonCommand{'C'}; die unless defined $Command;
	    my $Title=$Command;
	    if ($ThisButtonCommand{'T'}){$Title=$ThisButtonCommand{'T'};}; 
	    my $Picture=$ThisButtonCommand{'P'};
	    if ($Picture){
		push @ThisButtonDataStructure,
		[
		 Button => '',
		 -image => $M->Photo(-file => "$::CodeDir/Bitmaps/$Picture"),
		 -command => [\&PaletteChoice,$Command]
		 ]
		 } else {
		     push @ThisButtonDataStructure,
		     [Button => $Title,
		      -command => [\&PaletteChoice,$Command]
		      ]
		      };
	}; # while NextLine
	$D->Menubutton
	    (-text => $ButtonText,
	     -underline => 0,
	     -width => 20,
	     -menuitems =>\@ThisButtonDataStructure)->pack(-side => 'top');
    }; # while AllButtonData    
    
    $D->resizable(0,0); # prevent resizing
    $D->transient($M);   
    $D->waitVisibility;

    return $self;
};
sub AreYouOK{
    # We don't even try to cover all the ways the user could possibly close the dialog. Instead, we check if the dialog widget is still existing before trying anything fromt the outside.
    if ($::Palette && Exists($::Palette)){return 1;}; 
    return 0;
};
sub PaletteChoice{
    my $NewCommand=shift or die;
    $::UIManager->PaletteChoice($NewCommand);
};
sub destroy{
    $::Palette->destroy;
    undef $::Palette;
};
1;
