package Thing;
use strict;
## 'GO' contains graphics object of command
## 'FULLGO' contains all the graphics of the thing.

my @RedundantEntries=('FULLGO','HEADGO', 'REDRAWDIRTY','NEWASCII','NEWASCIITYPE','NEWASCIILISPLINK');
sub SetLispObject{
    my $self=shift or die;
    my %P=@_;
    die unless $P{'TYPE'};
    die unless defined $P{'INDEX'};
    die unless defined $P{'PATH'};
    die unless defined $P{'OLD'};
    die unless defined $P{'NEW'};
    
    my $TypeIdentifier='LISPOBJECT';
  
    if ($P{'TYPE'} eq 'RESULT'){$TypeIdentifier='LISPOBJECT_RESULT'};
    
    my $OldLispobject=
	$self->{'COMMANDS'}->
	    [$P{'INDEX'}]->
	    {$P{'PATH'}}->
	    {$TypeIdentifier} or die;
    
    # The root object has to be quoted and the output from Maxima is quoted. 
    # But if we insert the Maxima output somewhere into our expression, we have to 
    # remove the quotation mark, otherwise there will be trouble.
    if($OldLispobject eq $P{'OLD'}){	
	$self->{'COMMANDS'}->
	    [$P{'INDEX'}]->
	    {$P{'PATH'}}->
	    {$TypeIdentifier}=$P{'NEW'};
    } else {
	delete $P{'NEW'}->{'QUOTED'};
	$OldLispobject->ReplaceObject('OLD'=>$P{'OLD'},'NEW'=> $P{'NEW'}) or die;
    };
    $self->SetDirty('EVAL','REDRAW');

    #Note, that this sub is not called in any case from MaximaController.
#    print "setlispobje-call\n";
    $::ThingManager->MakeEvaldirtyIfNeeded;
};
sub SetLispResult{  
    my $self=shift or die;
    my %P=@_;
    die unless defined $P{'INDEX'};
    die unless defined $P{'NEW'};
    $self->{'COMMANDS'}->
	[$P{'INDEX'}]->
	{''}->
	{'LISPOBJECT_RESULT'}=$P{'NEW'};

    # Careful: Just that one result has been set does not mean, that all the other results are up-to-date as well. We'll leave it anyway.
    $self->ClearEvaldirty;
};
sub SetDoc{
    my $self=shift or die;
    my $DocLine=shift;
    delete $self->{'DOCLINE'};
    if ($DocLine){
	$self->{'DOCLINE'}=$DocLine;
    }; 
    $self->SetDirty('REDRAW');

};
sub QueryDoc{
    my $self=shift or die;
    my $Doc=$self->{'DOCLINE'} or return '';
    return $Doc;
};
sub QueryEvaldirty{
    my $self=shift or die;
    if ($self->{'EVALDIRTY'}){return 1;} return 0;
};
sub QueryTeXResDirty{
    my $self=shift or die;
    if ($self->{'GENERATETEXRES'} && $self->{'TEXRESENABLED'}){return 1;} return 0;
};
sub QueryTeXResEnabled{
    my $self=shift or die;
    if ($self->{'TEXRESENABLED'}){return 1;} return 0;
};
sub QueryNoEval{
    my $self=shift or die;
    if ($self->{'NOEVAL'}){return 1;} return 0;
};
sub SetDirty{
    my $self=shift or die;
    my $RecallNoEval=$self->QueryFlag(FLAGMODE => 'THING',FLAG => 'RECALLNOEVAL');
    my $Flag;while ($Flag=shift){
	die unless($Flag eq 'EVAL' || $Flag eq 'REDRAW');
	if ($Flag eq 'EVAL' && !$RecallNoEval){
#	    $self->{'REDRAWDIRTY'}=1;
	    $self->{'EVALDIRTY'}='NEW';
	} elsif ($Flag eq 'REDRAW'){
	    $self->{'REDRAWDIRTY'}=1;
	} elsif ($Flag eq 'EVAL' && $RecallNoEval){     
	    # no error, that's OK.
	} else {die $Flag;}
    };
    # This sub is called, when a change is made in the editor.
};
sub ClearEvaldirty{
     my $self=shift or die;
     delete $self->{'EVALDIRTY'};
};
sub NewFromASCII{
    my $self={}; bless $self; 
    my %P=@_;
    my $X=$P{'X'}; die unless defined $X;
    my $Y=$P{'Y'}; die unless defined $Y;
    my $ID=$P{'ID'}; die unless defined $ID;
    my $NewThingString=$P{'ASCII'} or die;
    my $Doc=$P{'DOC'};
    
    my $LispobjectCommand;

    # A simple number can be constructed without invoking Maxima.
    if ($NewThingString=~/^\d+$/){
	$LispobjectCommand=LispList::NewFromString(STRING => $NewThingString);
    } else {
      $LispobjectCommand=MaximaController::Maxima_ASCII_to_LISP('ASCII' => $NewThingString);
    };
    my $LispobjectResult=undef;
    
    $self->{'COMMANDS'}=
	[    
	     # First modifier list has only one entry! Only '', no coordinates
	     { '' => 
	       {'LISPOBJECT'=>$LispobjectCommand,
		'SHOWCOMMAND_LINE' => 1,
		'SHOWRESULT_LINE' => 1,
		'LISPOBJECT_RESULT'=>$LispobjectResult		    
		}
	   }
	     ];
#,
	     # Second level modifier.
#	     {'' => {'LISPOBJECT'=> MaximaController::Maxima_ASCII_to_LISP('ASCII' => '%+1'),'LISPOBJECT_RESULT'=>MaximaController::Maxima_ASCII_to_LISP('ASCII' => 'a+b+c+1+test')},
#	       '1' => {'LISPOBJECT'=>MaximaController::Maxima_ASCII_to_LISP('ASCII' => '%+test'),'DX' => 50,'DY' => 50}
	 
    $self->{X}=$X;$self->{Y}=$Y;
    $self->{SHOWDOC}=1;
    if ($Doc){$self->{DOCLINE}=$Doc;}
    $self->{SHOWHEAD}=1;
    $self->{'TEXRESENABLED'}=$::TeXOutputAsDefault;
    $self->{REDRAWDIRTY}=1;
    $self->{EVALDIRTY}=1;
    $self->{ID}=$ID;
    $self->{SHOWCONTENTS}=1;
    return $self;
};
sub MoveTo{
    my $self=shift or die;
    my %P=@_;
    my $X=$P{'X'}; die unless defined $X;
    my $Y=$P{'Y'}; die unless defined $Y;
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    if ($InterfacePage){
	$self->{'INTERFACEX'}=$X;
	$self->{'INTERFACEY'}=$Y;
    } else {
	$self->{'X'}=$X;
	$self->{'Y'}=$Y;
    };
    $self->{'REDRAWDIRTY'}='ALL';
};
sub QueryX{
    my $self=shift or die;
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    if ($InterfacePage){return $self->{'INTERFACEX'}} else {return $self->{'X'}};

};
sub QueryY{
    my $self=shift or die;
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    if ($InterfacePage){return $self->{'INTERFACEY'}} else {return $self->{'Y'}};

};
sub QueryNewCommand{
    my $self=shift or die;
    return ($self->{'NEWASCII'});
};
sub QueryID{
    my $self=shift or die;
    my $ID=$self->{'ID'}; die unless defined $ID;
    return $ID;
};
sub QueryListOfReferenceIDs{
    my $self=shift or die;
    my @CommandList=@{$self->{COMMANDS}} or die;
    my $CommandIndex;
    my @RefList;
    for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){
	my $Command=$CommandList[$CommandIndex] or die;
	my @PathList=keys(%{$Command}) or die;
	while ($#PathList>=0){
	    my $Path=shift @PathList;
	    my $Lispobject=$Command->{$Path}->{'LISPOBJECT'} or die "$CommandIndex.$Path.";
	    push @RefList,@{$Lispobject->QueryListOfReferenceIDs};
	};
  
	# include Refs set on the popup window, Line->Refs
	my $RefsInEval=$self->QueryFlag
	    (FLAG => 'REFSINEVAL',FLAGMODE => 'LINE',INDEX => $CommandIndex);;
	if ($RefsInEval){
	    my @Refs=@{$RefsInEval};
	    my $Ref; foreach $Ref(@Refs){
		push @RefList,$Ref;
	    };
	};
    };
    return \@RefList;
};
sub Redraw{
    my $self=shift or die;
    my $ID=$self->{'ID'}; die unless defined $ID;
    my %P=@_;
    my $Printmode=$P{'PRINTMODE'}; 
    if ($P{FORCE}){$self->{REDRAWDIRTY}=1;};
    if ($P{'KILL'}){   
	$::GlobalCanvas->delete("ID$ID");
	return;
    };  
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    my $X=$self->{'X'};  die unless defined $X;
    my $Y=$self->{'Y'};  die unless defined $Y;    if ($InterfacePage && $self->{'INTERFACE'}){
	if (!defined $self->{'INTERFACEX'}){$self->{'INTERFACEX'}=$self->{'X'};};
	if (!defined $self->{'INTERFACEY'}){$self->{'INTERFACEY'}=$self->{'Y'};};
	$X=$self->{'INTERFACEX'};
	$Y=$self->{'INTERFACEY'};
    };

    my $Dirty=$self->{REDRAWDIRTY};
    my $DrawThisThing=(!$InterfacePage || $self->{INTERFACE});
    if ($Dirty){
	$::GlobalCanvas->delete("ID$ID");	
#	print "Thing redraw $ID";#.join(":",@_)."\n";
	
	# Read: If the thing is on the current page
	if ($DrawThisThing){
	    $self->DeleteAllGOs;
	    my $Tag="ID$ID";
	    # Draw command
	    my @GOList=();
	    my $Anything=0;
	    if ($self->{SHOWHEAD}){
		$Anything=1;
		push @GOList,$self->GenerateHeadGO;
	    }; # if showhead
	    
	    if ($self->{SHOWCONTENTS}){
		my @MainGOList=@{$self->GenerateMainGOList};
		push @GOList,@MainGOList;
		if ($#MainGOList >=0){$Anything=1;}
	    };
	    if ($self->{SHOWDOC} && defined $self->{DOCLINE}){
		push @GOList,GO::NewFromString('STRING' => $self->{DOCLINE},'LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",'WRAP' => 1,'STYLE' => 'LIGHTBLUE');
		$Anything=1;
	    };
	    # if everything is turned off, a '#' to grab the thing
	    if (!$Anything){
		push @GOList,GO::NewFromString('STRING' => "#",'LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE");
	    };
	    
	    my $ListGO=GO::NewGOList('TYPE' => 'YLIST','LISPLINK'=>'NONE','GOLIST'=>\@GOList) or die;
	    $self->{'FULLGO'}=$ListGO;
	    
	    $ListGO->Draw('X'=>$X,'Y'=>$Y,'MODE'=>'DRAW','TAG' => $Tag);
	    $ListGO->DeleteNullgeom('LOWERLEVELONLY' => 1);
	    
	    # Modifiers require, that the main-GOs are already drawn.
	    $self->DrawModifiers;	
	    # Error messages require, that the main-GOs are already drawn.
	    $self->DrawAllErrorMessages;	    
	    
	}; # if interface (if visible)  
	delete $self->{'REDRAWDIRTY'};
    }; # if dirty
    
    # Remove / Redraw evaldirty-indicator
    my $IndicatorTag="ID$ID EVALDIRTYBAR";
    $::GlobalCanvas->delete($IndicatorTag);
    if ($DrawThisThing && $self->{EVALDIRTY}){
	my $DrawGeom=$self->{FULLGO}->QueryDrawGeom;
	$::GlobalCanvas->createRectangle
	    ($DrawGeom->{XMIN}-5, # Go one dot to the left, 
	     $DrawGeom->{YMIN}, # so that the selection-frame does not
	     $DrawGeom->{XMIN}-1,   # overlap
	     $DrawGeom->{YMAX},
	     -tags => ["ID$ID",$IndicatorTag],
	     -fill => 'red');	 
    }; # if evaldirty
};
sub DrawModifiers{
    my $self=shift or die;
    my $ID=$self->{'ID'}; die unless defined $ID;
    # Draw non-toplevel 'modifiers'
    # They will only be drawn, if the result from the previous evaluation
    # has been drawn.
    # There are no modifiers for command index 0! 
    
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    my $CommandIndex;for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){ 
	if ($self->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX=>$CommandIndex,
	     FLAG=>'SHOWCOMMAND')){
	    
	    die unless $CommandList[$CommandIndex];
	    my %CommandsForThisLevel=%{$CommandList[$CommandIndex]};
	    my @AllPaths=keys(%CommandsForThisLevel);
	    # We start with the previous command...
	    my $LastCommand=$CommandList[$CommandIndex-1] or die;
	    my $ThisCommand=$CommandList[$CommandIndex] or die;
	    
	    if ($self->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX=>$CommandIndex-1,
		 FLAG=>'SHOWRESULT')){
		my $Path; foreach $Path(@AllPaths){
		    my $Command=$CommandsForThisLevel{$Path} or die;
		    if ($Path ne ''){ # toplevel has already been drawn
			
			my @PathArray=split(/,/,$Path);
			
			# Find the lisp object, that this modifier is attached to (using the path).
			# ... take the full lisp object...
			my $LastFullLispObject=$LastCommand->{''}->{'LISPOBJECT_RESULT'};
			my $LispObject;
			if ($LastFullLispObject){
			    # ... and apply the path to get the object that will be replaced by this modifier.
			    $LispObject=$LastFullLispObject->QueryObjectToPath(@PathArray);
			};
			
			my $ModifAbsX; my $ModifAbsY;
			my $Geom;
			my $LastFullGO=$LastCommand->{''}->{'GO_RESULT'};
			
			# If it's a new object, there is no result yet.
			# Therefore we can't draw it.
			my $GO;
			if ($LispObject && $LastFullGO){
			    
			    $GO=$LastFullGO->QueryGOFromLispLink('LISPLINK' => $LispObject);
			}; 
			my $AttachedToUnknownSource;
			if ($GO){
			    my $Geom=$GO->QueryDrawGeom or die;
			    $ModifAbsX=$Geom->{'XMIN'}+$Command->{'DX'};
			    $ModifAbsY=$Geom->{'YMIN'}+$Command->{'DY'};
			    
			    # Frame around target
			    $::GlobalCanvas->createRectangle
				($Geom->{'XMIN'},
				 $Geom->{'YMIN'},
				 $Geom->{'XMAX'},
				 $Geom->{'YMAX'},
				 -outline => 'white',
				 -tags => "ID$ID");
			    
			    # Connection line
			    $::GlobalCanvas->createLine
				(($Geom->{'XMIN'}+$Geom->{'XMAX'})/2,
				 ($Geom->{'YMIN'}+$Geom->{'YMAX'})/2,
				 $ModifAbsX,
				 $ModifAbsY, 
				 -tags => "ID$ID");
			} elsif ($LastFullGO) {	
			    # The modifier may be attached to an object, which does not exist anymore.
			    $Geom=$LastFullGO->QueryDrawGeom or die;
			    $ModifAbsX=$Geom->{'XMIN'}+$Command->{'DX'};
			    $ModifAbsY=$Geom->{'YMIN'}+$Command->{'DY'};

			    #
			    # Show, that something has gone wrong
			    #
			    
			    $AttachedToUnknownSource=1;
			} else {die "Panic in Thing.pm...\n";}; # this case is impossible
			; # if not lispobject
			if (defined($ModifAbsX)){
			    #my $ModifierLispobject=$Command->{'LISPOBJECT'} or die;
			    my $ModifGO=$self->GetCommandGO
				(INDEX => $CommandIndex,
				 PATH => $Path);
			    
			    my $ModifGeom=$ModifGO->Draw('X'=>$ModifAbsX,'Y'=>$ModifAbsY,'MODE'=>'DRAW','TAG' => "ID$ID");	
			    if ($AttachedToUnknownSource){

				my $Off; for ($Off=-1;$Off<=1;$Off+=2){
				    $::GlobalCanvas->createRectangle
					($ModifGeom->{'XMIN'}-$Off,
					 $ModifGeom->{'YMIN'}-$Off,
					 $ModifGeom->{'XMAX'}-$Off,
					 $ModifGeom->{'YMAX'}-$Off,
					 -outline => 'red',
					 -tags => "ID$ID");
				};
			    };
			    
			    $Command->{'GO'}=$ModifGO;
			};
			
			}; # if not toplevel
		}; # foreach path
	    } else { # if SHOWRESULT on last result
		
		# If the result is not shown, but there are modifiers,
		# then the modifiers are not visible.
		# The user may forget, that they are hidden.
		# Therefore, draw a red line over this command to indicate,
		# that there are more modifiers.
		# Try to find any graphobject (depends on what part is
		# turned on)
		if($#AllPaths>0){
		    my $ThisGO=$ThisCommand->{''}->{'GO'};
		    
		    # if we haven't drawn this command, we can't be here
		    die unless $ThisGO;
		    
		    my $Geom=$ThisGO->QueryDrawGeom or die;
		    $::GlobalCanvas->createLine
			($Geom->{'XMIN'},$Geom->{'YMIN'},
			 $Geom->{'XMAX'},$Geom->{'YMIN'},
			 -fill => 'red',
			 -tags => "ID$ID");
		}; # if there are modifiers
	    }; # if not SHOWRESULT on last result
	}; # if not SHOWCOMMAND on this line
    }; # for CommandIndex	  
}; 
sub DrawAllErrorMessages{
    my $self=shift or die;
    my %P=@_;
    my $ID=$self->{ID}; die unless defined $ID;
    
    my @CommandList=@{$self->{COMMANDS}} or die;
    my $CommandIndex;for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){ 
	if ($self->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX=>$CommandIndex,
	     FLAG=>'SHOWCOMMAND')){
	    
	    die unless $CommandList[$CommandIndex];
	    my %CommandsForThisLevel=%{$CommandList[$CommandIndex]};
	    my @AllPaths=keys(%CommandsForThisLevel);
	    my $Command=$CommandList[$CommandIndex] or die;
	    
	    my $Path; foreach $Path(@AllPaths){
		if ($Command->{$Path}->{ERROR}){	
		    my $ThisFullGO=$Command->{$Path}->{'GO'} or die;
		    my $Geom=$ThisFullGO->QueryDrawGeom or die;
		    my $ErrorX=$Geom->{'XMIN'}-5; # leave some space for the red evaldirty bar
		    my $ErrorY=($Geom->{'YMIN'}+$Geom->{'YMAX'})/2;
		    
		    $::GlobalCanvas->createText
			($ErrorX,$ErrorY,
			 -text => $Command->{$Path}->{ERROR},
			 -anchor => 'e',-tags => "ID$ID",-fill => 'red');
		    # highlight fatal error with a big ugly star
		    if ($Command->{$Path}->{ERROR}=~/\<FATAL\>/){
			my $r=1000;
			my $i; for ($i=10;$i<190; $i+=20){
			    my $irad=$i*3.1415/180;
			    $::GlobalCanvas->createLine
				($ErrorX-$r*cos($irad),
				 $ErrorY-$r*sin($irad),
				 $ErrorX+$r*cos($irad),
				 $ErrorY+$r*sin($irad),
				 -tags => "ID$ID",
				 -fill => 'red');
			}; # for i
		    }; # if fatal error
		}; # if error in command
	    }; # foreach path
	}; # if SHOWCOMMAND
    }; # for commandindex
}; 
    
sub QueryCoordinatesWithin{
    my $self=shift or die;
    my %P=@_;
    my $GO=$self->{'FULLGO'} or die;
    return $GO->QueryCoordinatesWithin(%P);
};
sub QueryCommandGO{
    my $self=shift or die;
    return $self->{'COMMANDGO'};
};
sub QueryFullGO{
    my $self=shift or die;
    die unless $self->{'FULLGO'};
    return $self->{'FULLGO'};
};
sub PurgeRedundantData{
    my $self=shift or die;
    $self->DeleteAllGOs;
#    my $Entry; foreach $Entry(@RedundantEntries){
#	delete $self->{$Entry};
#    } 
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    my $Command;
    foreach $Command(@CommandList){
	my @Paths=keys(%{$Command});
	my $Path;foreach $Path(@Paths){
	    if ($Command->{$Path}->{'LISPOBJECT'}){
		$Command->{$Path}->{'LISPOBJECT'}->PurgeRedundantData;
	    };
	    if ($Command->{$Path}->{'LISPOBJECT_RESULT'}){
		$Command->{$Path}->{'LISPOBJECT_RESULT'}->PurgeRedundantData;
	    };
	};
    };	
};
sub QueryDefinition{
    my $self=shift or die;
    return $self->{'DEFINITION'};
};
sub QueryFlag{
    my $self=shift or die;
    my %P=@_;
    my $Mode=$P{FLAGMODE} or die;
    my $Flag=$P{FLAG} or die;
    my $Index=$P{INDEX};
    my $Path=$P{PATH};
    $self->CheckFlagDirtyMode(%P); # error checking
    my $Ret;
    if ($Mode eq 'THING'){
	$Ret=$self->{$Flag};
    } elsif ($Mode eq 'LINE'){
	die unless defined $Index;
	my $Command=$self->{COMMANDS}->[$Index]->{''} or die $Flag;
	$Ret=$Command->{$Flag.'_LINE'};
    } elsif ($Mode eq 'COMMAND'){
	die unless defined $Index; die $Flag unless defined $Path;
	my $Command=$self->{COMMANDS}->[$Index]->{$Path};
	$Ret=$Command->{$Flag.'_COMMAND'};
    } else {die;};
    if ($Ret){return $Ret;};
    return ''; # prevent undef, '' is better than 0 (could be a string)
};
sub SetFlag{
    my $self=shift or die;
    my %P=@_;
    my $New=$P{NEW}; die unless (defined $New || $P{TOLERATEINVALID});
    my $Mode=$P{FLAGMODE} or die;
    my $Flag=$P{FLAG} or die;
    my $Index=$P{INDEX};
    my $Path=$P{PATH};
    my $DoNotDirty=$P{DONOTDIRTY};

    if (!defined $New){$New='';};
    $New=~s/^\s*//g; # remove leading whitespace
    $New=~s/\s*$//g; # remove trailing whitespace
    my $DirtyMode=$self->CheckFlagDirtyMode(%P);
    my $Old=$self->QueryFlag(%P);
    # Check, if something changes
    my $OldString=$Old;
    my $NewString=$New;

    # For example REFSINEVAL is an array. Compare the contents.
    if (ref($Old) eq 'ARRAY'){$OldString=join(",",@{$Old})};
    if (ref($New) eq 'ARRAY'){$NewString=join(",",@{$New})};
    if ($Flag eq 'REFSINEVAL'){
	
	# Special rule here: If you read REFSINEVAL,
	# it is an array. But if you set it, it's a string.
	
	$New=~s/\s+//g;
	my $TestString=$New;
	$TestString=~s/\d+,//g;
	return unless ($TestString=~/^\d+$/ || $TestString eq '');
	
	my @Array=split(/,/,$New);
	$New=\@Array;
	}

    if ($OldString ne $NewString){
	if ($Mode eq 'THING'){
	    if ($New){
		$self->{$Flag}=$New;
	    } else {
		delete $self->{$Flag};
	    };
	} elsif ($Mode eq 'LINE'){
	    die unless defined $Index;
	    my $Command=$self->{COMMANDS}->[$Index]->{''} or die;
	    if ($New){
		$Command->{$Flag.'_LINE'}=$New;
	    } else {
		delete $Command->{$Flag.'_LINE'};
	    };
	} elsif ($Mode eq 'COMMAND'){
	    die unless defined $Index; die unless defined $Path;
	    my $Command=$self->{COMMANDS}->[$Index]->{$Path} or die "$Index $Path";
	    if ($New){
		$Command->{$Flag.'_COMMAND'}=$New;	
	    } else {
		delete $Command->{$Flag.'_COMMAND'};
	    };
	} else {die;};
	if (!$DoNotDirty){
	    if ($DirtyMode eq 'EVAL'){
		$self->SetDirty('EVAL','REDRAW');
	    };
	    $self->SetDirty('REDRAW');
	    
	    
	    if ($Flag eq 'RECALLNOEVAL' && $New){
		# The RecallNoEval-flag is set.
		# The user has decided to keep the old result, and he'll have his reasons.
		# So we turn off the Evaldirty flag for this expression, because the
		# output is what the user wants it to be
		$self->ClearEvaldirty;
	    } elsif ($Flag eq 'RECALLNOEVAL' && !$New){	
		$self->SetDirty('EVAL','REDRAW');
	    };
	}; # if flag changes
    }; # if not DoNotDirty

};
sub CheckFlagDirtyMode{
    my $self=shift or die;
    my %P=@_;
    my $Flag=$P{FLAG}; die unless defined $Flag;
    my $Mode=$P{FLAGMODE}; die $Flag unless defined $Mode;
    my %Flaglist=(
		  SHOWHEAD_THING => 'REDRAW',
		  SHOWDOC_THING => 'REDRAW',
		  SHOWCONTENTS_THING=> 'REDRAW',
		  INTERFACE_THING => 'REDRAW',
		  DEFINITION_THING => 'EVAL',
		  RECALLNOEVAL_THING => 'REDRAW',
		  EVALATLOAD_THING => 'REDRAW',
		  CIRCULAR_THING => 'REDRAW',
		  RECALLNOEVAL_LINE => 'EVAL',  # migrates at load
		  SHOWCOMMAND_LINE => 'REDRAW', # migrates at load
		  SHOWRESULT_LINE => 'REDRAW',  # migrates at load
		  NODISPFORM_LINE => 'EVAL',    # migrates at load
		  SKIPEVAL_LINE => 'EVAL',      # migrates at load
		  DONOTSIMP_LINE => 'EVAL',     # migrates at load
		  NUMER_LINE => 'EVAL',         # migrates at load
		  EVAL_LINE => 'EVAL',          # migrates at load
		  REFSINEVAL_LINE => 'EVAL',    # migrates at load
		  SEL_SELECT_COMMAND => 'EVAL',
		  SEL_UNSELECT_COMMAND => 'EVAL',
		  SEL_PICK_COMMAND => 'EVAL',
		  SEL_SELSIMP_LINE => 'EVAL',
		  UNIT_MULTIPLY_COMMAND => 'EVAL',
		  UNIT_CONVERT_COMMAND => 'EVAL',
		  GNUP_CREATEDATA_LINE => 'EVAL',
		  GNUP_INDEP_LINE => 'EVAL',
		  GNUP_MIN_LINE => 'EVAL',
		  GNUP_MAX_LINE => 'EVAL',
		  GNUP_PTS_LINE => 'EVAL',
		  GNUP_SMOOTH_LINE => 'EVAL',
		  GNUP_CURVETITLE_LINE => 'EVAL',
		  GNUP_PLOTDATA_LINE => 'EVAL',
		  GNUP_CURVELOG_LINE => 'EVAL',
		  GNUP_DIAGLOGX_LINE => 'EVAL',
		  GNUP_DIAGLOGY_LINE => 'EVAL',
		  GNUP_DIAGTITLE_LINE => 'EVAL',
		  GNUP_DIAGGRIDX_LINE => 'EVAL',
		  GNUP_DIAGGRIDY_LINE => 'EVAL',
		  GNUP_DIAGCOLOUR_LINE => 'EVAL',
		  GNUP_DIAGSCALE_LINE => 'EVAL',
		  GNUP_DIAGLABELX_LINE => 'EVAL',
		  GNUP_DIAGLABELY_LINE => 'EVAL',
		  GNUP_DIAGRANGEX_LINE => 'EVAL',
		  GNUP_DIAGRANGEY_LINE => 'EVAL',
		  GNUP_LISTPLOT_LINE => 'EVAL'
		  );
    die $Flag."_".$Mode unless $Flaglist{$Flag.'_'.$Mode};
    return $Flaglist{$Flag.'_'.$Mode};
};

sub QueryWhatsClicked{
    my $self=shift or die;
    my %P=@_;
    my $FullGO=$self->{'FULLGO'} or die;
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    if ($InterfacePage && !$self->{'INTERFACE'}){return undef;}

    # Check, if head was clicked
    if ($self->{'HEADGO'}){
	if ($self->{'HEADGO'}->QueryGraphObjectFromCoordinates(%P)){
	    return {'THING' => $self,
		    'FULLGOGEOM' => $self->{'FULLGO'}->QueryGeom,
		    'FULLGODRAWGEOM' => $self->{'FULLGO'}->QueryDrawGeom,
		    'DEPTH' => 99,
		    'TYPE' => 'HEAD'
		};
	};
    };
    my $Type;
    my $UpperLevelLispObject;
    my $CommandIndex;
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    my $Mode; foreach $Mode('MODIFIER','REST'){
  AllGOLoop:for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){
      my %CommandsForThisLevel=%{$CommandList[$CommandIndex]};
      my @PathList=keys(%CommandsForThisLevel);
      my $Path; foreach $Path(@PathList){
	  
	  # First try all non-'' objects:
	  # If a modifier is overlapping a command / result, you can move it away.
	  # This does not work the other way round, since the modifier will move with 
	  # the thing.
	  if ($Mode eq 'MODIFIER' && $Path ne '' || $Mode eq 'REST' && $Path eq ''){
	  my $Command=$CommandsForThisLevel{$Path} or die;

	  my $GOKey; foreach $GOKey('GO','GO_RESULT','GO_HANDLE'){ 
	      # we are doing extra loops for non-'' modifiers here, but that doesn't matter
	      # (A non-'' modifier cannot contain a GO_RESULT)
	      my $GO=$Command->{$GOKey};
	      if ($GO){
		  my $ClickedGOData=$GO->QueryGraphObjectFromCoordinates(%P);
		  if ($ClickedGOData){
		      if ($GOKey eq 'GO' && $Path eq ''){
			  $Type='COMMAND';
			  $UpperLevelLispObject=$Command->{'LISPOBJECT'} or die;
		      } elsif ($GOKey eq 'GO' && $Path ne ''){
			  $Type='MODIF';
			  $UpperLevelLispObject=$Command->{'LISPOBJECT'} or die;
		      } elsif ($GOKey eq 'GO_RESULT'){
			  $Type='RESULT';
			  $UpperLevelLispObject=$Command->{'LISPOBJECT_RESULT'} or die;
		      } elsif ($GOKey eq 'GO_HANDLE'){
			  $Type='HANDLE';
		      };
		      my $ReturnData={
			  'GOGEOM' => $GO->QueryGeom,
			  'GODRAWGEOM' => $GO->QueryDrawGeom,
			  'THING' => $self,
			  'TYPE' => $Type,
			  'INDEX' => $CommandIndex,
			  'PATH' => $Path,
			  'FULLGOGEOM' => $FullGO->QueryGeom,
			  'FULLGODRAWGEOM' => $FullGO->QueryDrawGeom,
			  'DEPTH' => $ClickedGOData->{'DEPTH'}
		      };
		      
		      my $Lispselection=$ClickedGOData->{'OBJECT'}->QueryLispLink;
		      if (ref($Lispselection)){ # NONE doesn't count
			  $ReturnData->{LISPSELECTION}=$Lispselection;
			  $ReturnData->{UPPERLEVELLISPOBJECT}=$UpperLevelLispObject;
		      };
		      return $ReturnData;
		  }; # if ClickedGOData
	      }; # if GO
	  }; # for GOKey
      }; # if right order (check modifiers first)
      }; # foreach path
  }; # for CommandIndex
}; # for mode

    # Finally, we check if the full graphobject was clicked somewhere.
    # This is necessary to be able to select an expression, whose display
    # is completely disabled (there will be only a # character to click on).
    if ($self->{'FULLGO'}){
	if ($self->{'FULLGO'}->QueryGraphObjectFromCoordinates(%P)){
	    return {'THING' => $self,
		    'FULLGOGEOM' => $self->{'FULLGO'}->QueryGeom,
		    'FULLGODRAWGEOM' => $self->{'FULLGO'}->QueryDrawGeom,
		    'DEPTH' => -1,
		    'TYPE' => 'HEAD'
		    };
	};
    }; # if FullGO


};
sub HighlightPreciseSelection{
    my $self=shift or die;
    my %P=@_;
    my $Type=$P{'SELECTIONTYPE'} or die;
    my $Sel=$P{'LISPSELECTION'}; die unless $Sel || ($Type eq 'HEAD' || $Type eq 'HANDLE');
    my $Index=$P{'INDEX'}; die unless defined $Index || ($Type eq 'HEAD');
    my $Path=$P{'PATH'}; die unless defined $Path || ($Type eq 'HEAD');
    my $GOToHighlight;
    my $GOMode;
    if ($Type eq 'COMMAND' || $Type eq 'MODIF'){
	$GOMode='GO';
    } elsif ($Type eq 'RESULT'){
	$GOMode='GO_RESULT';
    } elsif ($Type eq 'HEAD'){
	$GOMode='FULL';
    } elsif ($Type eq 'HANDLE'){
	$GOMode='GO_HANDLE';
    } else {die;} 
    if ($GOMode eq 'FULL'){
	$GOToHighlight=$self->{'FULLGO'} or return 0; #might be on other page...
	$GOToHighlight->HighlightPreciseSelection('MODE' => 'CURSOR','FORCE' => 1);
    } elsif ($GOMode eq 'GO_HANDLE'){
	$GOToHighlight=$self->{'COMMANDS'}->[$Index]->{''}->{$GOMode};
	if ($GOToHighlight){ # will not be there, if the command / result has just been turned on again (editormode remains the same).
	    $GOToHighlight->HighlightPreciseSelection('FORCE' => 1,'MODE' => 'CURSOR');
	};
    } else {
	$GOToHighlight=$self->{'COMMANDS'}->[$Index]->{$Path}->{$GOMode} or return 0; # might be on other page, deleted...
	$GOToHighlight->HighlightPreciseSelection('LISPLINK' => $Sel,'MODE' => 'CURSOR');
    };
    if ($GOToHighlight){
	return 1;
    }; 
    return 0;
};

sub MoveModifierRelative{
    my $self=shift or die; 
    my %P=@_;    
    my $Path=$P{'PATH'}; die unless defined $Path;
    my $Type=$P{'TYPE'} or die;
    my $Index=$P{'INDEX'} or die;
    my $DX=$P{'DX'}; die unless defined $DX;
    my $DY=$P{'DY'}; die unless defined $DY;
    my $Command=$self->{'COMMANDS'}->[$Index]->{$Path} or die;
    $Command->{'DX'}+=$DX;
    $Command->{'DY'}+=$DY;
    $self->{'REDRAWDIRTY'}='ALL';
};
sub QueryLispToplevelCommand{
    my $self=shift or die; 
    my $a=$self->{'COMMANDS'}   # All commands
    [0]                         # toplevel command
    {''}                        # Modifier path
    ->{'LISPOBJECT'} or die;    # Lisp object
    return $a;
};
sub QueryCommands{
    my $self=shift or die; 
    return $self->{'COMMANDS'};
};
sub QuerySortedPaths{
    my $self=shift or die; 
    my %P=@_;
    my $Index=$P{'INDEX'}; die unless defined $Index;
    my @SortedPathList=();

    my $Command=$self->{'COMMANDS'}[$Index] or die;
    my @PathList=keys(%{$Command}) or die;

    my %PathListSorted=();
    
    # List to hash
    while ($#PathList>=0){
	my $Path=shift @PathList;
	my $Length=0;
	my $Path2=$Path;
	while($Path2=~s/,//){$Length++;}
	if (!$PathListSorted{$Length}){$PathListSorted{$Length}=[];}
	push @{$PathListSorted{$Length}},$Path;
    };
    my @AllLengths=keys(%PathListSorted);
    @AllLengths=sort @AllLengths;
    my $Length;
    my @AllPaths=();
    foreach $Length(@AllLengths){
	push @AllPaths,@{$PathListSorted{$Length}};
    };
    if ($P{'SHORTEST_FIRST'}){} elsif ($P{'LONGEST_FIRST'}){@AllPaths=reverse(@AllPaths)} else {die;}
    return \@AllPaths;
};
sub SetError{ 
    my $self=shift or die; 
    my %P=@_;
    my $Index=$P{INDEX}; die unless defined $Index;
    my $Path=$P{PATH}; die unless defined $Path;
    my $Error=$P{ERROR}; die unless defined $Error;
    # If the present error message ends exactly with the new error message, just ignore it.
    # Use quotemeta, because otherwise []$^ would be interpreted as regexp
    my $OldError=$self->{COMMANDS}->[$Index]->{$Path}->{ERROR};
    if (!defined $OldError){$OldError='';};
    my $ErrorCopy=$Error;
    $ErrorCopy=~s/\s//g; # remove whitespace
    $OldError=~s/\s//g; 
    my $ErrorQuoteMeta=quotemeta($ErrorCopy);

    return if $OldError=~/$ErrorQuoteMeta/i;
    # Append otherwise.
    $self->{COMMANDS}->[$Index]->{$Path}->{ERROR}.=$Error;
};
sub ClearErrors{
    my $self=shift or die; 
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    my $CommandIndex;
    for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){
	my $Command=$CommandList[$CommandIndex] or die;
	my @PathList=keys(%{$Command}) or die;
	while ($#PathList>=0){
	    my $Path=shift @PathList;
	    delete $Command->{$Path}->{'ERROR'};
	};
    };
};
sub GenerateHeadGO{
    my $self=shift or die;
    my $ID=$self->{'ID'}; die unless defined $ID;
    my @Headline=();   
    $self->{'HEADGO'}=GO::NewFromString('STRING' => " ID $ID ",'LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",'STYLE'=>'BOX');
    push @Headline,$self->{'HEADGO'};
    # Command is dirty?
    my @Dirty=();
#    if ($self->QueryEvaldirty){ push @Dirty,'Result';};
    if ($self->QueryFlag
	(FLAGMODE => 'THING',FLAG => 'CIRCULAR')){ push @Dirty,'Circular';};
#    };
    if ($#Dirty>-1){
	push @Headline,GO::NewFromString
	    ('STRING' => " ".join("/",@Dirty)." ",
	     'LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",'STYLE'=>'DARKRED');
    }
    if ($self->QueryFlag
	(FLAGMODE => 'THING',FLAG => 'DEFINITION')){	
	push @Headline,GO::NewFromString('STRING' => ' Definition ','LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",'STYLE'=>'GREENBOX');	
    };
    if ($self->QueryFlag
	(FLAGMODE => 'THING',FLAG => 'EVALATLOAD')){ 
	push @Headline,GO::NewFromString
	    ('STRING' => ' AutoExec ','LISPLINK'=>"NONE",
	     'PARENTOPERATOR'=>"NONE",'STYLE'=>'GREENBOX');	};
    if ($self->QueryFlag
	(FLAGMODE => 'THING',FLAG => 'RECALLNOEVAL')){ 
	push @Headline,GO::NewFromString
	    ('STRING' => ' Recall only ',
	     'LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",
	     'STYLE'=>'DARKRED');	};
    return GO::NewGOList('TYPE' => 'XLIST','LISPLINK'=>'NONE','GOLIST'=>\@Headline) or die;
};
sub GenerateMainGOList{
    my $self=shift or die;
    my @GOList;
    my $CommandIndex;
    my @CommandList=@{$self->{'COMMANDS'}} or die;

    for ($CommandIndex=0;$CommandIndex<=$#CommandList;$CommandIndex++){
	die unless $CommandList[$CommandIndex];
	my $Anything;

	my %CommandsForThisLevel=%{$CommandList[$CommandIndex]};

	my $ShowCommand=$self->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX => $CommandIndex,
	     FLAG => 'SHOWCOMMAND'); 
	my $ShowResult=$self->QueryFlag
	    (FLAGMODE => 'LINE',
	     INDEX => $CommandIndex,
	     FLAG => 'SHOWRESULT'); 

	my $ToplevelCommand=$CommandsForThisLevel{''} or die;
	my $Lispcommand=$ToplevelCommand->{'LISPOBJECT'} or die;
	my $Lispresult=$ToplevelCommand->{'LISPOBJECT_RESULT'};

	if ($ShowCommand && $Lispcommand){

	    # This will generate modifiers applied through the
	    # properties dialog
	    my $CommandGO=$self->GetCommandGO
		(INDEX => $CommandIndex,
		 PATH => '');
	    
#$Lispcommand->GetGO('STYLE' => 'COMMAND') or die;
	    push @GOList,$CommandGO;
	    $ToplevelCommand->{'GO'}=$CommandGO;
	    $Anything=1;
	};
	if ($ShowResult && $Lispresult){
	    
	    # If the result has been modified, highlight.
	    my $Style='RESULT';
	    if ($self->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $CommandIndex,
		 FLAG => 'RECALLNOEVAL')){
		$Style='DARKRED';
	    };
	    my $ResultGO=$Lispresult->GetGO('STYLE' => $Style) or die;
	    push @GOList,$ResultGO;
	    $ToplevelCommand->{'GO_RESULT'}=$ResultGO;
	    $Anything=1;
	};
	# is there anything at all visible for this line=
	if (!$Anything){
	    # No? Then create a placeholder to be able to select it.
	    my $Placeholder=GO::NewFromString('STRING' => '...','LISPLINK'=>"NONE",'PARENTOPERATOR'=>"NONE",'WRAP' => 1);
	    push @GOList,$Placeholder;
	    # This is neither command nor result.
	    # Let's call it a handle:
	    $ToplevelCommand->{'GO_HANDLE'}=$Placeholder;
	};
    }; # for CommandIndex
    
    return \@GOList;
};	
sub DeleteAllGOs{
    my $self=shift or die;
    delete $self->{'HEADGO'}; 
    delete $self->{'FULLGO'}; 
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    my $Command;
    foreach $Command(@CommandList){
	my @Paths=keys(%{$Command});
	my $Path;foreach $Path(@Paths){
	    delete $Command->{$Path}->{GO};
	    delete $Command->{$Path}->{GO_RESULT};
	    delete $Command->{$Path}->{GO_HANDLE};
	};
    };	
};
sub DeleteCommand{
    my $self=shift or die;
    my %P=@_;
    my $Index=$P{'INDEX'}; die unless defined $Index;
    my @CommandList=@{$self->{'COMMANDS'}} or die;
    die if $#CommandList <0;
    return if $#CommandList==0;
    my @Commands=@{$self->{'COMMANDS'}};
    splice(@Commands,$Index,1,());
    $self->{'COMMANDS'}=\@Commands;
    $self->SetDirty('EVAL','REDRAW');
};
sub GetCommandGO{
    my $self=shift or die;
    my %P=@_;
    my $Index=$P{INDEX}; die unless defined $Index;
    my $Path=$P{PATH}; die unless defined $Path;
    my $Lispobject=$self->{COMMANDS}->[$Index]->{$Path}->{LISPOBJECT} or die;
    my $Protect=$self->QueryFlag
	(FLAGMODE => 'COMMAND',
	 INDEX => $Index,
	 PATH => $Path,
	 FLAG => 'SEL_SELECT');
    my $GO;
    if (!$::UIManager->QueryFeature('FEATURE' => 'SELECTION')){
	$Protect=0;};
    if (!$Protect){
	$GO=$Lispobject->GetGO('STYLE' => 'COMMAND') or die;
    } else {
	my $GroupLispobject=LispList::NewFromString
	    ('STRING' => "\$$Protect",'UNREAL' => 1);
	my $ProtectedArg=LispList::NewUnreal
	    (OPERATOR => '$SEL',
	     'DATA' => [$Lispobject,$GroupLispobject]);
	$GO=$ProtectedArg->GetGO('STYLE' => 'COMMAND','LISPLINK' =>$Lispobject ) or die;
    }; 
    
    # Create markers for all features, which are used.
    # Their Lisplink is set to the command. So clicking them will return the command.
    my %InvisibleFeatures=%{$self->QueryInvisibleFeatures(%P)};
    if (%InvisibleFeatures){
	my @FeatureMarkersList=();
	my $Featurename; foreach $Featurename(keys(%InvisibleFeatures)){
	    my $FeatureMarkerGO=GO::NewFromString
		('STRING' => $Featurename,
		 'LISPLINK'=>$Lispobject,
		 'PARENTOPERATOR'=>"NONE",
		 'STYLE' => 'BOX');
	    push  @FeatureMarkersList,$FeatureMarkerGO;
	};
	my $FeatureMarkersGO;
	if ($#FeatureMarkersList>=0){
	    if ($#FeatureMarkersList==0){
		$FeatureMarkersGO=$FeatureMarkersList[0];
	    } else {
		# stack feature markers on top of each other
		$FeatureMarkersGO=GO::NewGOList('TYPE' => 'YLIST','LISPLINK'=>$Lispobject,'GOLIST'=>\@FeatureMarkersList) or die;
	    };
	    # and display that stack on the left side of the expression
	    $GO=GO::NewGOList('TYPE' => 'XLIST','LISPLINK'=>$Lispobject,'GOLIST'=>[$FeatureMarkersGO,GO::NewSeparator(' ',$Lispobject),$GO]);
	};
    }; # if invisible features
    return $GO;
};
sub QueryInvisibleFeatures{
    my $self=shift or die;
    my %P=@_;
    my $Index=$P{INDEX}; die unless defined $Index;
    my $Path=$P{PATH}; die unless defined $Path;
    my %ReturnValue=();
    my $UsesIt;
    my $Flag; 

    $UsesIt=0;foreach $Flag('SEL_UNSELECT','SEL_PICK'){
	if ($self->QueryFlag
	    (FLAGMODE => 'COMMAND',
	     INDEX => $Index,
	     PATH => $Path,
	     FLAG => $Flag)){$UsesIt=1;}
    
    };
    if ($UsesIt){
	$ReturnValue{' s '}=1;};

    if ($Path eq ''){ # flag only applies to the line, not to modifiers
	$UsesIt=0;foreach $Flag('EVAL','NUMER','DONOTSIMP','SKIPEVAL'){
	    if ($self->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => $Flag)){$UsesIt=1;}
	};
	if ($UsesIt){
	    $ReturnValue{' e '}=1;
	};
    };
    if ($Path eq ''){ # flag only applies to the line, not to modifiers
	$UsesIt=0;foreach $Flag('GNUP_CREATEDATA','GNUP_PLOTDATA'){
	    if ($self->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => $Flag)){$UsesIt=1;}
	};
	if ($UsesIt){
	    $ReturnValue{' g '}=1;
	};
    };
    if ($Path eq ''){ # flag only applies to the line, not to modifiers
	$UsesIt=0;foreach $Flag('SEL_SELSIMP'){
	    if ($self->QueryFlag
		(FLAGMODE => 'LINE',
		 INDEX => $Index,
		 FLAG => $Flag)){$UsesIt=1;}
	};
	if ($UsesIt){
	    $ReturnValue{' s '}=1;
	};
    };
    $UsesIt=0;foreach $Flag('UNIT_MULTIPLY','UNIT_CONVERT'){
	if ($self->QueryFlag
	    (FLAGMODE => 'COMMAND',
	     INDEX => $Index,
	     PATH => $Path,
	     FLAG => $Flag)){$UsesIt=1;}
    };
    if ($UsesIt){
	$ReturnValue{' u '}=1;
    };
    return \%ReturnValue;
};
#sub DESTROY{print "t";}
1;
__END__
