package ThingManager;
use strict;
#use diagnostics;
use Symaxx2::Thing;
use Symaxx2::MySubs;
# use Symaxx2::FreezeThaw qw(freeze thaw);
sub New{
    my $self={'THINGLIST' => [],'NEXTID'=>0}; bless $self; return $self;
};
sub QueryThing{
    my $self=shift or die;
    my %P=@_;
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $TolerateInvalid=$P{'TOLERATEINVALID'};
    my $Thing;
    my $ID=$P{'ID'}; 
    if (defined $ID){
	foreach $Thing(@{$Thinglist}){
	    if ($Thing->QueryID eq $ID){return $Thing;};
	};
	if ($TolerateInvalid){return undef;}
	die "unknown ID: $ID";
    };
    my $X=$P{'X'};
    my $Y=$P{'Y'};
    if (defined $X && defined $Y){
  	foreach $Thing(@{$Thinglist}){
	    if ($Thing->QueryCoordinatesWithin('X'=>$X,'Y'=>$Y)){return $Thing;};
  	};
	return undef;
    };
    die 'no proper command given';
};
sub AddFromASCII{
    my $self=shift or die;
    my %P=@_;
    my $Thinglist=$self->{'THINGLIST'}; die unless defined $Thinglist;
    my $NewID=++$self->{'NEXTID'};
    my $NewThing=Thing::NewFromASCII(%P,'ID' => $NewID);
    push @{$Thinglist},$NewThing;
    return $NewThing;
};
sub QueryListOfNewCommandIDs{
    # returns list with ID as key and new command as value
    my $self=shift or die;
    my %ReturnList=();
    my $Thing; foreach $Thing(@{$self->{'THINGLIST'}}){
	if (defined $Thing->QueryNewCommand){
	    die unless $Thing->{'ID'};
	    $ReturnList{$Thing->{'ID'}}=$Thing->{'NEWASCII'};
	};
    };
    return \%ReturnList;
}; 
sub QueryListOfResultsForNextEvaluation{
    # which results have to be present, before the next evaluation
    # can be done?

    my $self=shift or die;
    my %P=@_;
    my $TeXOnly=$P{'TEXONLY'};
    my $AllRefsList={}; 
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $Thing; foreach $Thing(@{$Thinglist}){
	my $ThingID=$Thing->QueryID;
	# if a thing is set as NOEVAL, then its references are of no interest
	# here (because it does not get evaluated).
	# However, it may be referenced from other expressions.
	# We simply remove the evaldirty-flag, if it's set NOEVAL.
	# Then it does not get evaluated.
	# Switching off the flag will set evaldirty-status again.
	if ($Thing->QueryNoEval){delete $Thing->{'EVALDIRTY'};};
	
	# We don't handle dirty definitions here. They are instead
	# generated by QueryDefinitions.
	# This must be run also in TeX mode.

	if (!$Thing->QueryDefinition){
	    if (!$TeXOnly){
		if ($Thing->QueryEvaldirty){
		    $AllRefsList->{$ThingID}++;
		    my @AllReferences=@{$Thing->QueryListOfReferenceIDs};
		    my $ID; foreach $ID(@AllReferences){
			if ($self->QueryThing('ID' => $ID,'TOLERATEINVALID' => 1)){
			    $AllRefsList->{$ID}++; # Creates new entry, if needed.
#		    print "$ThingID references $ID\n";
			};
		    };
		}; # if thing is evaldirty
		# include the reference to create a TeX-Result always
	    }; # if not TeXOnly
	    if ($Thing->QueryTeXResDirty){
#	    print "*** New eval for tex: $ThingID***\n";
		$AllRefsList->{$ThingID}++; # Creates new entry, if needed.
	    };
	}; # if not definition   
    }; # for each thing
    
    my @Array=keys(%{$AllRefsList});
    return \@Array;
};
sub QueryDefinitions{
    my $self=shift or die;
    # possible FIXME: the order is arbitrary. In the case f(g(x)) it seens not to matter, in which
    # order f(x) and g(x) are defined.

    # If a definition is set as NOEVAL, it is not generated.
    my @FunDefs=();
    my $Thing; foreach $Thing(@{$self->{'THINGLIST'}}){
	if ($Thing->QueryDefinition && !$Thing->QueryNoEval){
	    push @FunDefs,$Thing->QueryID;
	};
    };
    return \@FunDefs;
};
sub ListOfDefinitionsWhichUseReferences{
    my $self=shift or die;
    my @Errorlist=();
    my $Thing; foreach $Thing(@{$self->{'THINGLIST'}}){
	if ($Thing->QueryDefinition){
	    my @AllReferences=@{$Thing->QueryListOfReferenceIDs};
	    if ($#AllReferences !=-1){push @Errorlist,$Thing->{'ID'};};
	};
    }; # foreach Thing
    return \ @Errorlist
};
sub MakeEvaldirtyIfNeeded{
#    print "TM-EvaldirtyIfNeeded\n";
    my $self=shift or die;
    my $Thinglist=$self->{THINGLIST} or die;
    
    # Is one of the definitions dirty? Then the whole mess gets dirty.
    # One point for sophistication, but ten for robustness :-)
    my $SetAllDirty=0;
    my $Thing; 
  CheckLoop:foreach $Thing(@{$Thinglist}){      
      if ($Thing->QueryDefinition && $Thing->QueryEvaldirty){
	  $SetAllDirty=1;last CheckLoop;
      };
  };
    if ($SetAllDirty){
	foreach $Thing(@{$Thinglist}){
	    if (!$Thing->QueryDefinition && !$Thing->QueryEvaldirty){
		# Note that we don't set 'REDRAW' here. 
		$Thing->SetDirty('EVAL');
	    };
	};
	return;
    };
    my %CleanThings; 
    foreach $Thing(@{$Thinglist}){
	if (!$Thing->QueryEvaldirty){
	    $CleanThings{$Thing->QueryID}=$Thing;
	};
    };
    my $Change=1;my $ID;
    ChangeLoop:while($Change){
	$Change=0;
	my @L=keys(%CleanThings);
	foreach $ID(@L){
	    $Thing=$CleanThings{$ID} or die;
	    if (!$Thing->QueryEvaldirty){
		my @AllReferences=@{$Thing->QueryListOfReferenceIDs};
		my $Ref; foreach $Ref(@AllReferences){
		    # reference to existing thing?
		    my $ReferencedThing=$self->QueryThing(ID => $Ref,TOLERATEINVALID => 1);
		    if (!$ReferencedThing || $ReferencedThing->QueryEvaldirty){
			# Note that we don't set 'REDRAW' here. That's the trick with
			# the evaldirty-indicator: Don't redraw all the referencing
			# things, if the referenced thing gets evaldirty.
			$Thing->SetDirty('EVAL');
			$Change=1;
			delete $CleanThings{$ID};
			next ChangeLoop;
		    } # if referenced thing dirty or referenced thing does not exist
		}; # foreach referenced thing
	    } # if this thing is still clean
	    else {
		delete $CleanThings{$Thing};
	    };
	}; # foreach Thing we've got
	# Everything is OK - quit
    }; # while Change
};
sub QueryIDList{
    my $self=shift or die;
    my $AllRefsList={}; 
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $Thing; foreach $Thing(@{$Thinglist}){
	$AllRefsList->{$Thing->QueryID}++;
    };
    my @Array=keys(%{$AllRefsList});
    return \@Array;
};
sub Delete{
    my $self=shift or die;
    my %P=@_;
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $ThingToKill=$P{'THING'} or die;

    # Before deleting, we make sure, that everything, which references to the
    # deleted object, is set dirty. This also works fine for definitions (if a 
    # def is deleted, everything is set evaldirty).
    $ThingToKill->SetDirty('EVAL');
    $self->MakeEvaldirtyIfNeeded;

    my @NewThingList=();
    my $Thing; foreach $Thing(@{$Thinglist}){
	if ($Thing ne $ThingToKill){
	    push @NewThingList,$Thing;
	} else {
	    # we have to remove the graphics object, because it gets deleted now.
	    # Otherwise it will stay on the screen forever.
	    $Thing->Redraw('KILL' => 1);
	};
    };
    $self->{'THINGLIST'}=\@NewThingList;
    $self->MakeEvaldirtyIfNeeded;

    # Essential to avoid memory leakage
    $ThingToKill->PurgeRedundantData; 
    $self->Redraw('ALL' => 1);
};
sub Save{  
    my $self=shift or die;
    my %P=@_;
    my $Thinglist=$self->{'THINGLIST'} or die;
    
    my $OutFilename=$P{'FILENAME'}; die unless defined $OutFilename;
    open(Filehandle,">$OutFilename") or die "unable to open $OutFilename for write";
    $::Filename=MySubs::SplitAbsolutePath($OutFilename)->{'FILE'} or die;
    $::WorkDir=MySubs::SplitAbsolutePath($OutFilename)->{'PATH'} or die;
    
    $self->PurgeRedundantData;
    my %SaveData=('THINGLIST' => $self->{'THINGLIST'},
		  'NEXTID' => $self->{'NEXTID'}, 
		  'MAXIMAUSEUNITS' => $::MaximaUseUnits,
		  'MAXIMAUSESELECTION' => $::MaximaUseSelection,
		  'MAXIMAUSEGNUPLOT' => $::MaximaUseGnuplot,
		  'MAXIMAUSEFORMAT' => $::MaximaUseFormat,
		  'MAXIMAUSENUMERICAL' => $::MaximaUseNumerical,
		  'SYMAXX/2VERSION' => $::Revision);
    # If you want to debug saved data, set INDENT to '' instead of undef.
    if ($::UseFreezeThaw){
	my $String=FreezeThaw::freeze(%SaveData);
	print Filehandle $String;
    } else {
      MySubs::SaveAsPerlCode('FILEHANDLE' => *Filehandle,'INDENT' => undef,'STRUCTURE'=>\%SaveData);
    };
    
    close(Filehandle);  
    $self->RecreateRedundantThingData;
    $::UIManager->SetTitle;
    $::MainDirty=0;
};
sub Load{
    my $self=shift or die;
    my %P=@_;
    die unless defined $P{'FILENAME'};
    my $InFilename=$P{'FILENAME'};
    open(Filehandle,"<$InFilename") or die "$InFilename could not be opened for reading.";
  
    my %LoadData; 
    my @LoadFile=<Filehandle>;
    close Filehandle;
    my $LoadString=join('',@LoadFile);
    # Avoid memory leak
    $self->PurgeRedundantData;
    $LoadString=~/^(...)/ or die;
    if ($1 eq 'FrT'){
	die;
	%LoadData=FreezeThaw::thaw($LoadString);
	#print keys(%Res)."\n";
#	%LoadData=%{$Res};
    } else {    
	# The load mechanism is a good example of applied perl virtues: 
	# Saving simply generates perl code, which is assigned to a variable by evaluating in the next line.
	# Slim, efficient, but smells funny :-)
	
	eval("%LoadData=%{$LoadString;}") or die "Loading of data failed. Please make sure that $InFilename is a valid symaxx-file.";
	
    };
    my $LoadFileVersion='---unknown---';
    if(defined($LoadData{'SYMAXXVERSION'})){die "This is a 'Symaxx' (1)-File. Please use that program!";};
    if(defined($LoadData{'SYMAXX/2VERSION'})){$LoadFileVersion=$LoadData{'SYMAXX/2VERSION'}};;
    if ($LoadFileVersion ne $::Revision){
	print "The file was saved with Symaxx/2 $LoadFileVersion, while this is Symaxx/2 $::Revision. When you save, your file will be saved in $::Revision-format. For files saved with 0.07 or earlier, the 'alias' functionality is no more supported. Otherwise, there are no known compatibility problems.\n";
#	if ($LoadFileVersion < 0.10){
#	    print "###NOTE:###\nThe postprocessing of Maxima output is now done in Maxima instead of Symaxx (for example, turn x^(-1*a) into 1/a).\nThis will cause your old files to look funny. Evaluate everything once to go back to normal.\nHint: Setting and clearing the DEFINITION flag on any expression will reevaluate the whole file.";
#	}
    };
    if ($LoadFileVersion < 0.10){	
	return if (UIManager::YesOrNo('TITLE' => 'New file format','TEXT'=> "You are loading a file in an older file format.\nIf you save, you won't be able to use the older version anymore.\nPlease save under a different filename!\nPress Yes to load, No to abort") ne 'YES'); 
    };
    
    # Retrieve all the stuff
    $self->{'THINGLIST'}=MySubs::UndefAlt($LoadData{'THINGLIST'},{}); # empty thinglist is OK
    
    # Convert Pre-0.10-code by moving the flags to the new locations / names.
    my $Thing; my @Thinglist=@{$self->{THINGLIST}};
    my @T=('FLAGMODE' => 'THING','TOLERATEINVALID' => 1);
    my @L=('FLAGMODE' => 'LINE', 'INDEX' => 0,'TOLERATEINVALID' => 1);

    # The EPSFILE functionality is no more supported. 
    # Use EPS("file",scale).
    # But print a note, if one of the old pictures is deleted.
    my @Deleted_EPS_Pictures=();
    foreach $Thing(@Thinglist){
	if ($Thing->{LISPCOMMAND}){
	    $Thing->{COMMANDS}->[0]->{''}->{LISPOBJECT}=$Thing->{LISPCOMMAND};delete $Thing->{LISPCOMMAND};
	    # Any old thing *must* have a command - otherwise the file is not valid. So we do the other flags here, too.

	    $Thing->SetFlag(@T,FLAG => 'SHOWHEAD', NEW => $Thing->{SHOWHEAD});
	    delete $Thing->{SHOWHEAD};
	    delete $Thing->{PROTECT};
#	    $Thing->SetFlag(@T,FLAG => 'SHOWDOC',NEW => $Thing->{SHOWDOC});delete $Thing->{SHOWDOC};
	    $Thing->SetFlag(@T,FLAG => 'SHOWCONTENTS',NEW => 1); # new

	    # Migrate those flags, which were earlier an attribute of the whole 'Thing'
	    my $FlagMovesToLine;
	    foreach $FlagMovesToLine('SHOWCOMMAND','SHOWRESULT','EVAL','NUMER','RECALLNOEVAL','NODISPFORM','SKIPEVAL','DONOTSIMP'){
		$Thing->SetFlag
		    (@L,
		     FLAG => $FlagMovesToLine,
		     NEW => $Thing->{$FlagMovesToLine},
		     'DONOTDIRTY' => 1 # if it changes, do not set dirty
		     ); 
		delete $Thing->{$FlagMovesToLine}; 
	    };
	    
	    
	    if ($Thing->{REFSINEVAL}){
		# REFSINEVAL has to be set as a string
		$Thing->SetFlag(@L,FLAG => 'REFSINEVAL',NEW => join(",",@{$Thing->{REFSINEVAL}}),
		     'DONOTDIRTY' => 1 # if it changes, do not set dirty
		     );
		delete $Thing->{REFSINEVAL};
	    };
	};
	if (defined($Thing->{LISPRESULT})){
	    $Thing->{COMMANDS}->[0]->{''}->{LISPOBJECT_RESULT}=$Thing->{LISPRESULT};
	    delete $Thing->{LISPRESULT};
	};

	delete $Thing->{'TEXRESENABLED'};
	if (defined $Thing->{EPSFILE} || defined $Thing->{EPSSCALE}){
	    push @Deleted_EPS_Pictures,$Thing->{ID};
	    delete $Thing->{EPSFILE}; delete $Thing->{EPSSCALE};
	}
    };

    if ($#Deleted_EPS_Pictures>-1){
	print "*** Warning ***\nThe loaded file includes objects, which display an EPS picture.\n";
	print "Objects in question: ID ".join(",",@Deleted_EPS_Pictures)."\n";
	print "This functionality is no more supported. Please include an EPS(\"filename\",Scale)\n";
	print "statement instead, which will do the same.\n\n";
    };
    
    $self->{NEXTID}=MySubs::UndefAlt($LoadData{NEXTID},0);
    $::MaximaUseUnits=MySubs::UndefAlt($LoadData{MAXIMAUSEUNITS},0);
    $::MaximaUseSelection=MySubs::UndefAlt($LoadData{MAXIMAUSESELECTION},0);
    $::MaximaUseGnuplot=MySubs::UndefAlt($LoadData{MAXIMAUSEGNUPLOT},0);
    $::MaximaUseFormat=MySubs::UndefAlt($LoadData{MAXIMAUSEFORMAT},0);
    # Numeric: If not mentioned in file, turn on (introduced later for example
    # for LinInterp
    $::MaximaUseNumerical=MySubs::UndefAlt($LoadData{MAXIMAUSENUMERICAL},1);
    
    $::Filename=MySubs::SplitAbsolutePath($InFilename)->{'FILE'} or die;
    $::WorkDir=MySubs::SplitAbsolutePath($InFilename)->{'PATH'} or die;
    $::UIManager->SetTitle;

    # Some things may create their own EPS-pictures. In this case, the
    # user will set the switch GENERATETEXRES, which results in an automatical
    # evaluation at loading time.
    my @EvalOrder=();
    foreach $Thing(@Thinglist){
	$Thing->PurgeRedundantData; # do this at save AND load to be sure
	if ($Thing->{'EVALATLOAD'}){
	    $Thing->SetDirty('EVAL','REDRAW');
	    push @EvalOrder,$Thing->{'ID'};
	};
    };
    $self->MakeEvaldirtyIfNeeded;
    if ($#EvalOrder >=0){
	print "Evaluating, because the ID(s) ".join(",",@EvalOrder)." are marked to run at loading time\n";
	&MaximaController::Evaluate;
    };
    # recreate all redundant data at the next redraw
    $self->Redraw('ALL' => 1,'FORCE' => 1); 
    $::MainDirty=0;
    
}; 
sub Redraw{
    my $self=shift or die;
    my %P=@_;
    my @IDList;
    my $Force=$P{'FORCE'}; 
    if ($P{'IDLIST'}){@IDList=@{$P{'IDLIST'}};};
    if ($P{'ALL'}){@IDList=@{$self->QueryIDList};};
    my $Printmode=$P{'PRINTMODE'}; 
    if ($Printmode){$::PrintMode=1} else {$::PrintMode=0};
    my $Arrows=0;
    my $ID; foreach $ID(@IDList){	
	my $Thing=$self->QueryThing('ID' => $ID);
	$Thing->Redraw('FORCE' => $Force,'PRINTMODE' => $Printmode); 
    };
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    $::GlobalCanvas->delete('ARROW');

    # Draw arrows (not on interface page)
    if (!$InterfacePage){
	@IDList=@{$self->QueryIDList};
	foreach $ID(@IDList){	
	    my $Thing=$self->QueryThing('ID' =>$ID);
	    my @AllReferences=@{$Thing->QueryListOfReferenceIDs};
	    my $Ref; foreach $Ref(@AllReferences){
		my $TargetThing=$self->QueryThing('ID'=>$Ref,'TOLERATEINVALID' => 1);
		my $EditorID;
		if ($::UIManager->{'THINGINEDITOR'}){
		    $EditorID=$::UIManager->{'THINGINEDITOR'}->QueryID;
		};
		my $Color='lightblue';
		if ($TargetThing){
		    if ($::AllArrows || ($EditorID && ($Thing->QueryID == $EditorID ||($TargetThing->QueryID == $EditorID))) ){
			$Color='blue';};
		    $::GlobalCanvas->createLine
			    ($Thing->{'X'}-10,$Thing->{'Y'}-10,
			     $TargetThing->{'X'}-10,$TargetThing->{'Y'}-10,
			     -arrow=>'first',-tags=>'ARROW',-fill=>$Color);
		};
	};
	};
    };

    # Adjust the scroll bars to fit the contents
    $::UIManager->RecalcCanvas;
    # delete the selection-cursor
    $::GlobalCanvas->delete('PRECISESELECTIONCURSOR');
    $::UIManager->ArrangeCanvasLevels;
};
sub ClearCircular{
    my $self=shift or die;	
    my $Thing;
    foreach $Thing(@{$self->{'THINGLIST'}}){
	$Thing->SetFlag('FLAGMODE' => 'THING','FLAG' => 'CIRCULAR', 'NEW' => 0);
    };
};
sub QueryWhatsClicked{
    my $self=shift or die;
    my %P=@_;
    my $X=$P{'X'}; die unless defined $X; 
    my $Y=$P{'Y'}; die unless defined $Y; 
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $Thing;my $ClickInfo;
    my $BestClickInfo=undef;
    my $InterfacePage=$::UIManager->QueryInterfacePage;
    foreach $Thing(@{$Thinglist}){
	if ($Thing->{'INTERFACE'} || !$InterfacePage){
	    $ClickInfo=$Thing->QueryWhatsClicked('X'=>$X,'Y'=>$Y);
	    if ($ClickInfo){
		# Click on head has highest priority.
		if (!$BestClickInfo || $BestClickInfo->{DEPTH}<$ClickInfo->{DEPTH}){
		    $BestClickInfo=$ClickInfo;
		};
	    }; # if not hit
	}; # if visible (interface page)
    }; # foreach thing
    return $BestClickInfo;
};
sub RecreateRedundantThingData{
    my $self=shift or die;
    my $Thing; foreach $Thing(@{$self->{'THINGLIST'}}){
	# Give each thing a chance to recreate its TeX output
	$Thing->{'GENERATETEXRES'}=1;
    };
#   # recreate TeX result
#   &MaximaController::Evaluate('TEXONLY' => 1);
    # recreate all redundant data 
    $self->Redraw('ALL' => 1,'FORCE' => 1);
};
sub InvalidateAll{
    my $self=shift or die;
    my $Thing;
    foreach $Thing(@{$self->{'THINGLIST'}}){
	$Thing->SetDirty('EVAL');
    };
    $self->Redraw('ALL' => 1,'FORCE' => 1);    
};
sub PurgeRedundantData{
    my $self=shift or die;
    # Delete all redundant data
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $Thing; foreach $Thing(@{$Thinglist}){
	$Thing->PurgeRedundantData;
    };
};
sub IsADefinitionDirty{
    my $self=shift or die;
    my $Thinglist=$self->{'THINGLIST'} or die;
    my $Thing; foreach $Thing(@{$Thinglist}){
	if ($Thing->QueryDefinition && $Thing->QueryEvaldirty){return 1;}
    }; 
    return 0;
};
1;
